#!/usr/bin/env python
""" MultiQC submodule to parse output from truvari bench """
import json
import logging
import os
import re

from multiqc.plots import bargraph, scatter

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"(\d+\.\d+\.[\d\.\-\w]+)"


class BenchSummary:
    def parse_bench_stats(self):
        """Find truvari bench logs and parse their data"""
        data = {}
        for f in self.find_log_files("truvari/bench"):
            inside_json_block = False
            json_text = ""
            version = None
            for line in f["f"].splitlines():
                if "Stats:" in line:
                    inside_json_block = True
                    json_text = "{\n"
                    continue

                if inside_json_block:
                    json_text += line
                    if line.startswith("}"):
                        inside_json_block = False

                # Get version from log
                # Log lines look like:
                #   2022-12-01 18:25:54,715 [INFO] Truvari v4.0.0.dev0+detached
                # or
                #   2022-08-25 00:31:16,781 [INFO] Truvari version: 3.5.0-dev
                if "Truvari" in line:
                    match = re.search(VERSION_REGEX, line)
                    if match:
                        version = match.group(1)

            if not json_text:
                log.warning(f"Could not find the 'Stats' JSON block in file: {f['fn']}")
                continue

            # Load stats
            try:
                stats = json.loads(str(json_text))
            except json.decoder.JSONDecodeError as e:
                log.debug(e)
                log.warning(f"Could not parse the 'Stats' JSON block in file: {f['fn']}")
                continue

            # Use output directory as sample name
            f["s_name"] = os.path.basename(f["root"])
            f["s_name"] = self.clean_s_name(f["s_name"], f, root=os.path.dirname(f["root"]))
            if f["s_name"] in data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")

            # Some stats were renamed in truvari 4.0.0 (commit 6e37058)
            # This renames them back to the old names for backwards compatibility
            if all(key in stats for key in ["TP-call_TP-gt", "TP-call_FP-gt", "TP-call", "call cnt"]):
                stats["TP-comp_TP-gt"] = stats["TP-call_TP-gt"]
                stats["TP-comp_FP-gt"] = stats["TP-call_FP-gt"]
                stats["TP-comp"] = stats["TP-call"]
                stats["comp cnt"] = stats["call cnt"]

            self.add_data_source(f, section="bench")
            data[f["s_name"]] = stats

            if version is not None:
                self.add_software_version(version, f["s_name"])

        # Filter to strip out ignored sample names
        data = self.ignore_samples(data)

        # Return if no samples
        if len(data) == 0:
            return len(data)

        # Write parsed report data to a file (restructure first)
        self.write_data_file(data, "multiqc_truvari_bench")

        # General Stats Table
        bench_headers = dict()
        bench_headers["precision"] = {
            "title": "Precision",
            "description": "Precision of the SV calls. Definition: TP-comp / (TP-comp + FP)",
            "suffix": "%",
            "modify": lambda x: x * 100,
            "placement": 100,
            "max": 100,
            "min": 0,
            "scale": "PRGn",
        }
        bench_headers["recall"] = {
            "title": "Recall",
            "description": "Recall of the SV calls. Definition: TP-base / (TP-base + FN)",
            "suffix": "%",
            "modify": lambda x: x * 100,
            "placement": 101,
            "max": 100,
            "min": 0,
            "scale": "BrBG",
        }
        bench_headers["f1"] = {
            "title": "F1",
            "description": "F1 score of the SV calls. Definition: 2 * ((Recall * Precision) / (Recall + Precision))",
            "suffix": "%",
            "modify": lambda x: x * 100,
            "placement": 102,
            "max": 100,
            "min": 0,
            "scale": "RdYlGn",
        }
        bench_headers["gt_concordance"] = {
            "title": "GT concordance",
            "description": "Genotype concordance. Definition: TP-comp with GT / (TP-comp with GT + TP-comp w/o GT)",
            "scale": "GnBu",
            "suffix": "%",
            "placement": 103,
            "max": 100,
            "min": 0,
            "hidden": True,
            "modify": lambda x: x * 100,
        }

        self.general_stats_addcols(data, bench_headers)

        # Make bar graph
        bar_data = [{}, {}, {}, {}]
        for sample, sample_data in data.items():
            # Comp
            bar_data[0][sample] = {
                "TP": sample_data["TP-comp"],
                "FP": sample_data["FP"],
            }
            # Base
            bar_data[1][sample] = {
                "TP": sample_data["TP-base"],
                "FN": sample_data["FN"],
            }
            # TP (comp) GT
            bar_data[2][sample] = {
                "Match": sample_data["TP-comp_TP-gt"],
                "No match": sample_data["TP-comp_FP-gt"],
            }
            # TP (base) GT
            bar_data[3][sample] = {
                "Match": sample_data["TP-base_TP-gt"],
                "No match": sample_data["TP-base_FP-gt"],
            }

        bar_categories = [
            # Comp
            {
                "TP": {"name": "TP", "color": "#648FFF"},
                "FP": {"name": "FP", "color": "#DC267F"},
            },
            # Base
            {
                "TP": {"name": "TP", "color": "#648FFF"},
                "FN": {"name": "FN", "color": "#FFB000"},
            },
            # TP (comp) GT
            {
                "Match": {"name": "Match", "color": "#785EF0"},
                "No match": {"name": "No match", "color": "#FE6100"},
            },
            # TP (base) GT
            {
                "Match": {"name": "Match", "color": "#785EF0"},
                "No match": {"name": "No match", "color": "#FE6100"},
            },
        ]

        bar_config = {
            "id": "truvari-bench-classifications-plot",
            "title": "Truvari: Classifications",
            "ylab": "Counts",
            "tt_suffix": " calls",
            "data_labels": [
                {"name": "Comp", "ylab": "Counts"},
                {"name": "Base", "ylab": "Counts"},
                {"name": "TP (comp) GT", "ylab": "Counts"},
                {"name": "TP (base) GT", "ylab": "Counts"},
            ],
        }

        self.add_section(
            name="Classifications",
            anchor="truvari-bench-classifications",
            description="Bargraph of SV call classifications parsed from the output from `truvari bench`",
            helptext="""
            Bargraph of SV call classifications from the perspectives of the *comp* and
            *base* ("truth") VCFs. Four different groups of calls are shown:

            #### Comp
            Compares TP and FP calls in the *comp* VCF relative the *base*. The
            classifications are:

            - **TP**: Comp call matches a base call
            - **FP**: Comp call does not match a base call

            #### Base
            Compares TP and FN calls in the *base* VCF relative the *comp*. The
            classifications are:

            - **TP**: Base call matches a comp call
            - **FN**: Base call does not match a comp call

            #### TP (comp) GT
            Compares TP calls in the *comp* VCF relative to the *base* VCF for
            matching genotypes. The classifications are:

            - **Match**: TP call in comp has matching genotype in base
            - **No match**: TP call in comp does not have matching genotype in base

            #### TP (base) GT
            Compares TP calls in the *base* VCF relative to the *comp* VCF for
            matching genotypes. The classifications are:

            - **Match**: TP call in base has matching genotype in comp
            - **No match**: TP call in base does not have matching genotype in comp

            For more information, see the [truvari bench wiki](https://github.com/acenglish/truvari/wiki/bench)
            """,
            plot=bargraph.plot(bar_data, bar_categories, pconfig=bar_config),
        )

        # Make scatter plot
        scatter_data = {}
        for i, (sample, sample_data) in enumerate(data.items()):
            scatter_data[sample] = {
                "x": sample_data["precision"] * 100.0,
                "y": sample_data["recall"] * 100.0,
            }

        scatter_config = {
            "id": "truvari-bench-pre-rec-plot",
            "marker_size": 5,
            "height": 560,  # increase height slightly to fit title.
            "ymax": 100,
            "ymin": 0,
            "xmax": 100,
            "xmin": 0,
            "xlab": "Precision (%)",
            "ylab": "Recall (%)",
            "square": True,
            "title": "Truvari: Precision vs. Recall",
            "tt_label": "{point.x:.1f}% precision<br/>{point.y:.1f}% recall",
        }
        self.add_section(
            name="Precision vs. Recall",
            anchor="truvari-bench-pre-rec",
            description="Precision vs. Recall for each sample. Parsed from the output of `truvari bench`",
            helptext="""
            Scatter plot of precision vs. recall comparing SV calls between two VCFs,
            one truth set ("base") and one to be evaluated ("comp"). The precision and
            recall values are calculated as follows:

            - **Precision**: TP-comp / (TP-comp + FP)
            - **Recall**: TP-base / (TP-base + FN)

            The TP, FP, and FN values are intrun defined as follows:

            - **TP (base)**: Number of matching calls from the base ('truth') VCF
            - **TP (comp)**: Number of matching calls from the comp VCF
            - **FP**: Number of non-matching calls from the comp VCF
            - **FN**: Number of non-matching calls from the base ('truth') VCF

            For more information, see the [truvari bench wiki](https://github.com/acenglish/truvari/wiki/bench)
            """,
            plot=scatter.plot(scatter_data, scatter_config),
        )

        # Return the number of logs that were found
        return len(data)



""" MultiQC module to parse output from Tophat """


import logging
import os
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Tophat",
            anchor="tophat",
            href="https://ccb.jhu.edu/software/tophat/",
            info="is a fast splice junction mapper for RNA-Seq reads. "
            "It aligns RNA-Seq reads to mammalian-sized genomes.",
            doi=["10.1186/gb-2013-14-4-r36", "10.1093/bioinformatics/btp120"],
        )

        # Find and load any Tophat reports
        self.tophat_data = dict()
        for f in self.find_log_files("tophat"):
            parsed_data = self.parse_tophat_log(f["f"])
            if parsed_data is not None:
                if f["s_name"] == "align" or f["s_name"] == "align_summary.txt":
                    s_name = os.path.basename(f["root"])
                else:
                    s_name = f["s_name"].split("align_summary.txt", 1)[0]
                s_name = self.clean_s_name(s_name, f)
                if s_name in self.tophat_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.add_data_source(f, s_name)
                self.tophat_data[s_name] = parsed_data

        # Filter to strip out ignored sample names
        self.tophat_data = self.ignore_samples(self.tophat_data)

        if len(self.tophat_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.tophat_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.tophat_data, "multiqc_tophat.txt")

        # Basic Stats Table
        self.tophat_general_stats_table()

        # Alignment Rate Plot
        self.tophat_alignment_plot()

    def parse_tophat_log(self, raw_data):
        """Parse the Tophat alignment log file."""

        if "Aligned pairs" in raw_data:
            # Paired end data
            regexes = {
                "overall_aligned_percent": r"([\d\.]+)% overall read mapping rate.",
                "concordant_aligned_percent": r"([\d\.]+)% concordant pair alignment rate.",
                "aligned_total": r"Aligned pairs:\s+(\d+)",
                "aligned_multimap": r"Aligned pairs:\s+\d+\n\s+of these:\s+(\d+)",
                "aligned_discordant": r"(\d+) \([\s\d\.]+%\) are discordant alignments",
                "total_reads": r"[Rr]eads:\n\s+Input\s*:\s+(\d+)",
            }
        else:
            # Single end data
            regexes = {
                "total_reads": r"[Rr]eads:\n\s+Input\s*:\s+(\d+)",
                "aligned_total": r"Mapped\s*:\s+(\d+)",
                "aligned_multimap": r"of these\s*:\s+(\d+)",
                "overall_aligned_percent": r"([\d\.]+)% overall read mapping rate.",
            }

        parsed_data = {}
        for k, r in regexes.items():
            r_search = re.search(r, raw_data, re.MULTILINE)
            if r_search:
                parsed_data[k] = float(r_search.group(1))

        # Exit if we didn't manage to parse enough fields - probably not a TopHat log
        # Note that Bowtie2 / HiSAT2 logs contain some but not all of these strings
        if len(parsed_data) < 4:
            return None

        parsed_data["concordant_aligned_percent"] = parsed_data.get("concordant_aligned_percent", 0)
        parsed_data["aligned_total"] = parsed_data.get("aligned_total", 0)
        parsed_data["aligned_multimap"] = parsed_data.get("aligned_multimap", 0)
        parsed_data["aligned_discordant"] = parsed_data.get("aligned_discordant", 0)
        parsed_data["unaligned_total"] = parsed_data["total_reads"] - parsed_data["aligned_total"]
        parsed_data["aligned_not_multimapped_discordant"] = (
            parsed_data["aligned_total"] - parsed_data["aligned_multimap"] - parsed_data["aligned_discordant"]
        )
        return parsed_data

    def tophat_general_stats_table(self):
        """Take the parsed stats from the Tophat report and add it to the
        basic stats table at the top of the report"""

        headers = {
            "overall_aligned_percent": {
                "title": "% Aligned",
                "description": "overall read mapping rate",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            },
            "aligned_not_multimapped_discordant": {
                "title": f"{config.read_count_prefix} Aligned",
                "description": f"Aligned reads, not multimapped or discordant ({config.read_count_desc})",
                "min": 0,
                "scale": "PuRd",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
        }
        self.general_stats_addcols(self.tophat_data, headers)

    def tophat_alignment_plot(self):
        """Make the HighCharts HTML to plot the alignment rates"""

        # Specify the order of the different possible categories
        keys = {
            "aligned_not_multimapped_discordant": {"color": "#437bb1", "name": "Aligned"},
            "aligned_multimap": {"color": "#f7a35c", "name": "Multimapped"},
            "aligned_discordant": {"color": "#e63491", "name": "Discordant mappings"},
            "unaligned_total": {"color": "#7f0000", "name": "Not aligned"},
        }

        # Config for the plot
        config = {
            "id": "tophat_alignment",
            "title": "Tophat: Alignment Scores",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        self.add_section(plot=bargraph.plot(self.tophat_data, keys, config))



""" MultiQC submodule to parse output from Picard CrosscheckFingerprints """

import logging
import re
from collections import OrderedDict, defaultdict
from csv import DictReader
from itertools import chain, groupby

from multiqc import config
from multiqc.plots import table, heatmap
from multiqc.utils.util_functions import strtobool

# Initialize the logger
log = logging.getLogger(__name__)

# This is a subset, the rest of the fields are self-descriptive
FIELD_DESCRIPTIONS = {
    "LEFT_SAMPLE": "The name of the left sample.",
    "LEFT_GROUP_VALUE": "The name of the left data-type group.",
    "RIGHT_SAMPLE": "The name of the right sample.",
    "RIGHT_GROUP_VALUE": "The name of the right data-type group.",
    "RESULT": "The categorical result of comparing the calculated LOD score against the threshold.",
    "DATA_TYPE": "The datatype used for the comparison.",
    "LOD_SCORE": "Log10 of the probability that the samples come from the same individual.",
    "LOD_SCORE_TUMOR_NORMAL": "LOD score with the assumption that Left is a Tumor.",
    "LOD_SCORE_NORMAL_TUMOR": "LOD score with the assumption that Right is a Tumor.",
    "LOD_THRESHOLD": "The LOD threshold used for this pairwise comparison.",
    "TUMOR_AWARENESS": "Whether this pairwise comparison was flagged for tumor awareness",
}


def parse_reports(module):
    """
    Find Picard CrosscheckFingerprints reports and parse their data.

    Stores the data in "Sample/Group - Sample/Group" groups since CrosscheckFingerprints
    does pairwise comparisons between samples at the level selected by `--CROSSCHECK_BY`.
    """

    row_by_number = dict()
    n_found_reports = 0

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/crosscheckfingerprints", filehandles=True):
        # Parse an individual CrosscheckFingerprints Report
        (metrics, comments) = _take_till(f["f"], lambda line: line.startswith("#") or line == "\n")
        header = next(metrics).rstrip("\n").split("\t")
        if "LEFT_GROUP_VALUE" not in header:
            # Not a CrosscheckFingerprints Report
            continue
        # Parse out the tumor awareness option and the lod threshold setting if possible
        tumor_awareness, lod_threshold = _parse_cli(comments[1])
        reader: DictReader = DictReader(metrics, fieldnames=header, delimiter="\t")
        lines = [
            (i, row)
            for i, row in enumerate(reader)
            if not module.is_ignore_sample(row["LEFT_SAMPLE"]) and not module.is_ignore_sample(row["RIGHT_SAMPLE"])
        ]
        if not lines:
            continue
        n_found_reports += 1
        for i, row in lines:
            # Clean the sample names
            row["LEFT_SAMPLE"] = module.clean_s_name(row["LEFT_SAMPLE"], f)
            row["LEFT_GROUP_VALUE"] = module.clean_s_name(row["LEFT_GROUP_VALUE"], f)
            row["RIGHT_SAMPLE"] = module.clean_s_name(row["RIGHT_SAMPLE"], f)
            row["RIGHT_GROUP_VALUE"] = module.clean_s_name(row["RIGHT_GROUP_VALUE"], f)

            row["RESULT"] = row["RESULT"].capitalize().replace("_", " ")

            # Set the cli options of interest for this file
            row["LOD_THRESHOLD"] = lod_threshold
            row["TUMOR_AWARENESS"] = tumor_awareness
            row_by_number[i] = row

            try:
                row["LOD_SCORE"] = float(row["LOD_SCORE"])
            except ValueError:
                row["LOD_SCORE"] = None

            # Add BEST_MATCH and BEST_MATCH_LOD if the result is unexpected or inconclusive
            if row["RESULT"].startswith("Unexpected") or row["RESULT"] == "Inconclusive":
                # Find the best match for the left sample
                left_sample = row["LEFT_SAMPLE"]
                best_match = max(
                    (r for j, r in lines if r["LEFT_SAMPLE"] == left_sample), key=lambda r: float(r["LOD_SCORE"])
                )
                row["BEST_MATCH"] = best_match["RIGHT_SAMPLE"]
                row["BEST_MATCH_LOD"] = float(best_match["LOD_SCORE"])

            module.add_data_source(f, section="CrosscheckFingerprints")

    # Only add sections if we found data
    if n_found_reports == 0:
        return 0

    # Sort the data by the left sample, then right sample
    row_by_number = OrderedDict(
        sorted(row_by_number.items(), key=lambda x: (x[1]["LEFT_SAMPLE"], x[1]["RIGHT_SAMPLE"]))
    )

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write data to file
    module.write_data_file(row_by_number, f"{module.anchor}_crosscheckfingerprints")

    # Add a per-sample table
    status_by_sample = dict()
    sorted_by_left_sample = sorted(row_by_number.values(), key=lambda r: r["LEFT_SAMPLE"])
    for left_sample, values in groupby(sorted_by_left_sample, key=lambda r: r["LEFT_SAMPLE"]):
        values = list(values)
        if values:
            status = "All expected"
            if all(v["RESULT"].startswith("Unexpected") for v in values):
                status = "All unexpected"
            elif any(v["RESULT"].startswith("Unexpected") for v in values):
                status = "Some unexpected"
            elif any(v["RESULT"] == "Inconclusive" for v in values):
                status = "Some inconclusive"
            status_by_sample[left_sample] = {"Crosschecks": status}
            best_match = max(values, key=lambda v: v["LOD_SCORE"])
            status_by_sample[left_sample]["Best match"] = best_match["RIGHT_SAMPLE"]
            status_by_sample[left_sample]["Best match LOD"] = best_match["LOD_SCORE"]

    sample_table_headers = {
        "Crosschecks": {
            "title": "Crosschecks",
            "description": "Flags if there are any unexpected or inconclusive matches.",
            "cond_formatting_rules": {
                "pass": [{"s_eq": "All expected"}],
                "warn": [{"s_eq": "Some inconclusive"}],
                "fail": [{"s_eq": "Unexpected"}, {"s_eq": "Some unexpected"}],
            },
        },
        "Best match": {
            "title": "Best match",
            "description": "The sample name with the highest LOD for this sample.",
        },
        "Best match LOD": {
            "title": "Best match LOD",
            "description": "The LOD score for the best matching sample.",
        },
    }

    def _order_status(status: str) -> int:
        # sorting order to show the bad results first
        if "all unexpected" in status.lower():
            return 0
        if "unexpected" in status.lower():
            return 1
        if "inconclusive" in status.lower():
            return 2
        return 3

    module.add_section(
        name="Crosscheck Fingerprints",
        anchor=f"{module.anchor}-crosscheckfingerprints-sample-table-section",
        description="Summary for each sample, showing the overall status along with the best matching sample name.",
        plot=table.plot(
            dict(sorted(status_by_sample.items(), key=lambda kv: _order_status(kv[1]["Crosschecks"]))),
            sample_table_headers,
            pconfig={
                "namespace": module.name,
                "id": f"{module.anchor}-crosscheckfingerprints-sample-table",
                "title": f"{module.name}: Crosscheck Fingerprints: Samples",
                "no_violin": True,
                "sortRows": False,
            },
        ),
    )
    for h in sample_table_headers.values():
        h["hidden"] = True
    module.general_stats_addcols(
        status_by_sample,
        sample_table_headers,
        namespace="CrosscheckFingerprints",
    )

    # Heatmap of the LOD scores for each pairwise comparison
    heatmap_data = defaultdict(dict)
    for row in row_by_number.values():
        left = row["LEFT_SAMPLE"]
        right = row["RIGHT_SAMPLE"]
        heatmap_data[left][right] = row["LOD_SCORE"]

    module.add_section(
        name="Crosscheck Fingerprints: Heatmap",
        anchor=f"{module.anchor}-crosscheckfingerprints-heatmap-section",
        description="Pairwise identity checking between samples and groups: heatmap of LOD scores.",
        plot=heatmap.plot(
            heatmap_data,
            xcats=list(heatmap_data.keys()),
            pconfig={
                "id": f"{module.anchor}-crosscheckfingerprints-heatmap",
                "title": f"{module.name}: Crosscheck Fingerprints",
                "zlab": "LOD score",
                "xcats_samples": True,
                "ycats_samples": True,
                "square": True,
                "legend": False,
                "reverseColors": True,
            },
        ),
    )

    # If the number of rows is > 100, do not show pairs with "Expected" status
    warning = ""
    if len(row_by_number) > 100:
        warning = (
            f"Note that there are too many pairwise comparisons to show in table "
            f" ({len(row_by_number)} > 100), so only unexpected or inconclusive pairs are shown."
        )
        row_by_number = {k: v for k, v in row_by_number.items() if not v["RESULT"].startswith("Expected")}

    module.add_section(
        name="Crosscheck Fingerprints: Pairwise Table",
        anchor=f"{module.anchor}-crosscheckfingerprints-table-section",
        description="Pairwise identity checking between samples and groups." + (f"<br>{warning}" if warning else ""),
        helptext="""
        Checks that all data in the set of input files comes from the same individual, based on the selected group granularity.
        """,
        plot=table.plot(
            data=dict(sorted(row_by_number.items(), key=lambda kv: (_order_status(kv[1]["RESULT"]), kv[0]))),
            headers=_get_table_headers(row_by_number),
            pconfig={
                "namespace": module.name,
                "id": f"{module.anchor}-crosscheckfingerprints-table",
                "title": f"{module.name}: Crosscheck Fingerprints",
                "save_file": True,
                "col1_header": "ID",
                "no_violin": True,
                "sortRows": False,
            },
        ),
    )

    return n_found_reports


def _take_till(iterator, fn):
    """
    Take from an iterator till `fn` returns false.

    Returns the iterator with the value that caused false at the front, and all the lines skipped till then as a list.
    """
    headers = []
    try:
        val = next(iterator)
        while fn(val):
            headers.append(val)
            val = next(iterator)
    except StopIteration:
        return ()

    return chain([val], iterator), headers


def _parse_cli(line):
    """Parse the Picard CLI invocation that is stored in the header section of the file."""
    tumor_awareness_regex = r"CALCULATE_TUMOR_AWARE_RESULTS(\s|=)(\w+)"
    lod_threshold_regex = r"LOD_THRESHOLD(\s|=)(\S+)"

    tumor_awareness = None
    lod_threshold = None

    tumor_awareness_match = re.search(tumor_awareness_regex, line)
    if tumor_awareness_match is not None:
        tumor_awareness = strtobool(tumor_awareness_match.group(2))

    lod_threshold_match = re.search(lod_threshold_regex, line)
    if lod_threshold_match is not None:
        lod_threshold = float(lod_threshold_match.group(2))

    return tumor_awareness, lod_threshold


def _get_table_headers(data_by_sample):
    """Create the headers config"""

    table_cols = [
        "RESULT",
        "DATA_TYPE",
        "LOD_THRESHOLD",
        "LOD_SCORE",
    ]
    table_cols_hidden = [
        "LEFT_RUN_BARCODE",
        "LEFT_LANE",
        "LEFT_MOLECULAR_BARCODE_SEQUENCE",
        "LEFT_LIBRARY",
        "LEFT_FILE",
        "RIGHT_RUN_BARCODE",
        "RIGHT_LANE",
        "RIGHT_MOLECULAR_BARCODE_SEQUENCE",
        "RIGHT_LIBRARY",
        "RIGHT_FILE",
        "DATA_TYPE",
        "LOD_THRESHOLD",
    ]

    # Allow customisation from the MultiQC config
    picard_config = getattr(config, "picard_config", {})
    table_cols = picard_config.get("CrosscheckFingerprints_table_cols", table_cols)
    table_cols_hidden = picard_config.get("CrosscheckFingerprints_table_cols_hidden", table_cols_hidden)

    # Add the Tumor/Normal LOD scores if any pair had the tumor_awareness flag set
    if any(row["TUMOR_AWARENESS"] for row in data_by_sample.values()):
        table_cols += [
            "LOD_SCORE_TUMOR_NORMAL",
            "LOD_SCORE_NORMAL_TUMOR",
        ]
    else:
        table_cols_hidden += [
            "LOD_SCORE_TUMOR_NORMAL",
            "LOD_SCORE_NORMAL_TUMOR",
        ]

    # Add Left and Right Sample names / groups, keeping it as minimal as possible
    def sample_group_are_same(x):
        return x["LEFT_SAMPLE"] == x["LEFT_GROUP_VALUE"] and x["RIGHT_SAMPLE"] == x["RIGHT_GROUP_VALUE"]

    if all(sample_group_are_same(values) for values in data_by_sample.values()):
        table_cols = [
            "LEFT_SAMPLE",
            "RIGHT_SAMPLE",
        ] + table_cols
        table_cols_hidden += ["LEFT_GROUP_VALUE", "RIGHT_GROUP_VALUE"]
    else:
        table_cols = [
            "LEFT_SAMPLE",
            "LEFT_GROUP_VALUE",
            "RIGHT_SAMPLE",
            "RIGHT_GROUP_VALUE",
        ] + table_cols

    headers = OrderedDict()
    for h in FIELD_DESCRIPTIONS:
        # Skip anything not set to visible
        if h not in table_cols:
            continue

        # Set up the configuration for the column
        h_title = (
            h.replace("_", " ")
            .strip()
            .lower()
            .capitalize()
            .replace("Lod score", "LOD")
            .replace("tumor normal", "T/N")
            .replace("normal tumor", "N/T")
        )
        headers[h] = {
            "title": h_title,
            "description": FIELD_DESCRIPTIONS.get(h),
            "namespace": "CrosscheckFingerprints",
            "scale": False,
        }

        # Rename Result to be a longer string so the table formats more nicely
        if h == "RESULT":
            headers[h]["title"] = "Match"
            headers[h]["cond_formatting_rules"] = {
                "pass": [{"s_contains": "Expected"}],
                "warn": [{"s_eq": "Inconclusive"}],
                "fail": [{"s_contains": "Unexpected"}],
            }

        # Add appropriate colors for LOD scores
        if h.startswith("LOD"):
            headers[h]["scale"] = "RdYlGn"
            headers[h]["shared_key"] = "LOD"
            headers[h]["bars_zero_centrepoint"] = True

        if h in table_cols_hidden:
            headers[h]["hidden"] = True

    # Add a new column showing the best match for the left sample if there are unexpected samples
    if "RESULT" in table_cols and "LEFT_SAMPLE" in table_cols and "RIGHT_SAMPLE" in table_cols:
        headers["BEST_MATCH"] = {
            "title": "Best match",
            "description": "The sample name with the highest LOD for the left sample.",
            "namespace": "CrosscheckFingerprints",
            "hidden": False,
        }
        headers["BEST_MATCH_LOD"] = {
            "title": "Best match LOD",
            "description": "The LOD score for the best matching sample for the left sample.",
            "namespace": "CrosscheckFingerprints",
            "hidden": False,
        }

    return headers



""" MultiQC module to parse output from VEP """


import ast
import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table
from multiqc.utils import mqc_colour

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """VEP"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="VEP",
            anchor="vep",
            href="https://www.ensembl.org/info/docs/tools/vep/index.html",
            info="Ensembl VEP determines the effect of your variants on genes, transcripts and protein sequences, "
            "as well as regulatory regions.",
            doi="10.1186/s13059-016-0974-4",
        )

        self.vep_data = dict()

        # Scan for VEP stats in plain html format
        for f in self.find_log_files("vep/vep_html", filehandles=True):
            self.parse_vep_html(f)
            self.add_data_source(f)

        # Scan for VEP stats in plain text format
        for f in self.find_log_files("vep/vep_txt", filehandles=True):
            self.parse_vep_txt(f)
            self.add_data_source(f)

        # Add version information
        for sample, data in self.vep_data.items():
            if "VEP run statistics" not in data:
                continue

            vep_version, api_version = data["VEP run statistics"]["VEP version (API)"].strip().split(" ")
            api_version = api_version.replace("(", "").replace(")", "")
            self.add_software_version(vep_version, sample)
            # Only add API version if it's different to VEP version
            if vep_version != api_version:
                self.add_software_version(api_version, sample, "VEP API")
        # Filter to strip out ignored sample names
        self.vep_data = self.ignore_samples(self.vep_data)

        # Stop if we didn't get any samples
        if len(self.vep_data) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.vep_data)} VEP summaries")

        # Write data to file
        self.write_data_file(self.vep_data, "vep")

        # Add general stats table
        self.add_stats_table()

        # Bar graphs
        self.bar_graph_variant_classes()
        self.bar_graph_consequences()
        self.bar_graph_sift()
        self.bar_graph_polyphen()
        self.bar_graph_variants_by_chromosome()
        self.bar_graph_position_in_protein()

    def extract_vep_html_data(self, chart_title, html_content):
        """Function for finding and extracting VEP stats that were stored as javascript arrays"""
        found_matches = re.findall(f"{chart_title}.*google.visualization.*;", html_content)
        if len(found_matches) > 0:
            array_content = re.search(r"\[\[.*]]", found_matches[0])
            if array_content:
                x = array_content.group(0)
                try:
                    rv = ast.literal_eval(x)
                    return rv
                except ValueError:
                    return None
            else:
                return None
        else:
            return None

    def parse_vep_html(self, f):
        """This Function will parse VEP summary files with HTML format"""
        # Initialise vep_data dictionary for the current sample
        if f["s_name"] not in self.vep_data:
            self.vep_data[f["s_name"]] = {}
        vepf = f["f"]
        html_content = vepf.read()
        # The tables with the titles given below have common format inside the javascript section
        titles = [
            "Variant classes",
            r"Consequences \(most severe\)",
            r"Consequences \(all\)",
            "Coding consequences",
            "SIFT summary",
            "PolyPhen summary",
            "Variants by chromosome",
            "Position in protein",
        ]
        # Grab values of tables with the titles given above
        for title in titles:
            # Remove backslashes which were required for re
            clean_title = title.replace("\\", "")
            extracted_data = self.extract_vep_html_data(title, html_content)
            if extracted_data:
                if clean_title not in self.vep_data[f["s_name"]]:
                    self.vep_data[f["s_name"]][clean_title] = {}
                plot_data = {}
                for i in range(1, len(extracted_data)):
                    plot_data[extracted_data[i][0]] = extracted_data[i][1]
                self.vep_data[f["s_name"]][clean_title] = plot_data

        # Finally parse general stats table which is an HTML table
        title = "General statistics"
        hits = re.search(r'General statistics</h3><table class="stats_table">.*?</table>', html_content)
        if hits:
            if title not in self.vep_data[f["s_name"]]:
                self.vep_data[f["s_name"]][title] = {}
            hit = hits.group(0)
            rows = re.findall(r"<tr>.*?</tr>", hit)
            for r in rows:
                cells = re.findall(r"<td>.*?</td>", r)
                if len(cells) == 2:
                    key = cells[0][4:-5]
                    value = cells[1][4:-5]
                    if value == "-":
                        continue
                    if key == "Novel / existing variants":
                        values = value.split("/")
                        novel = values[0].split("(")[0].replace(" ", "")
                        existing = values[1].split("(")[0].replace(" ", "")
                        self.vep_data[f["s_name"]][title]["Novel variants"] = int(novel)
                        self.vep_data[f["s_name"]][title]["Existing variants"] = int(existing)
                    else:
                        self.vep_data[f["s_name"]][title][key] = int(value)

    def parse_vep_txt(self, f):
        """This Function will parse VEP summary files with plain text format"""
        # Initialise vep_data dictionary for the current sample
        if f["s_name"] not in self.vep_data:
            self.vep_data[f["s_name"]] = {}
        vepf = f["f"]
        txt_data = {}
        title = ""
        for line in vepf.readlines():
            line = line.rstrip("\n")
            if len(line) < 1:
                continue
            if line[0] == "[":
                title = line.replace("[", "").replace("]", "")
                txt_data[title] = {}
                continue
            key, value = line.split("\t")
            if value == "-":
                continue
            if key == "Novel / existing variants":
                values = value.split("/")
                novel = values[0].split("(")[0].replace(" ", "")
                existing = values[1].split("(")[0].replace(" ", "")
                txt_data[title]["Novel variants"] = int(novel)
                txt_data[title]["Existing variants"] = int(existing)
            elif title != "VEP run statistics":
                txt_data[title][key] = int(value)
            else:
                txt_data[title][key] = value

        for title in txt_data.keys():
            if title not in self.vep_data[f["s_name"]]:
                self.vep_data[f["s_name"]][title] = {}
            self.vep_data[f["s_name"]][title] = txt_data[title]

    def add_stats_table(self):
        """Add a section with VEP General Statistics"""
        table_config = {
            "id": "vep-general-stats",
            "namespace": "VEP",
            "table_title": "VEP General Statistics",
        }
        table_data = {s_name: self.vep_data[s_name]["General statistics"] for s_name in self.vep_data}

        # Build the header configs
        cat_names = [
            "Overlapped regulatory features",
            "Overlapped transcripts",
            "Overlapped genes",
            "Existing variants",
            "Novel variants",
            "Variants filtered out",
            "Variants processed",
            "Lines of input read",
        ]
        # Set up the base config for each column
        table_cats = dict()
        color_list = ["Oranges", "Reds", "Blues", "Greens"]
        for order, header in enumerate(cat_names):
            table_cats[header] = {
                "name": header,
                "format": "{:,.0f}",
                "scale": color_list[order % 4],
            }
        # Column-specific customisation
        table_cats["Lines of input read"]["hidden"] = True
        table_cats["Variants processed"]["shared_key"] = "variants"
        table_cats["Variants filtered out"]["shared_key"] = "variants"
        table_cats["Novel variants"]["shared_key"] = "variants"
        table_cats["Existing variants"]["shared_key"] = "variants"

        self.add_section(
            name="General Statistics",
            anchor="vep-general-statistics",
            helptext="Table showing general statistics of VEP annotaion run",
            plot=table.plot(table_data, table_cats, table_config),
        )

    def bar_graph_variant_classes(self):
        title = "Variant classes"
        plot_data, plot_cats, plot_config = self._prep_bar_graph(title)
        htmlid = re.sub(r"\W+", "_", title).lower()
        if len(plot_data) == 0:
            return

        plot_config["ylab"] = "Number of variants"
        self.add_section(
            name=title,
            anchor=htmlid,
            description="Classes of variants found in the data.",
            plot=bargraph.plot(plot_data, plot_cats, plot_config),
        )

    def bar_graph_consequences(self):
        p_data = []
        p_cats = []
        p_config = {"data_labels": []}
        for title in ["Consequences (most severe)", "Coding consequences", "Consequences (all)"]:
            plot_data, plot_cats, plot_config = self._prep_bar_graph(title)
            p_data.append(plot_data)
            p_cats.append(plot_cats)
            p_config.update(plot_config)
            p_config["data_labels"].append(title)
        p_config["title"] = "VEP: Variant Consequences"
        p_config["ylab"] = p_config["data_labels"][0]

        if len(plot_data) == 0 or max([len(d) for d in plot_data]) == 0:
            return

        self.add_section(
            name="Consequences",
            anchor="consequences",
            description="Predicted consequences of variations.",
            plot=bargraph.plot(p_data, p_cats, p_config),
        )

    def bar_graph_sift(self):
        title = "SIFT summary"
        plot_data, plot_cats, plot_config = self._prep_bar_graph(title)
        htmlid = re.sub(r"\W+", "_", title).lower()
        if len(plot_data) == 0:
            return

        # Customise order and colours of categories
        p_cats = {
            "tolerated": {"color": "#59ae61"},
            "tolerated_low_confidence": {"color": "#a6db9f"},
            "deleterious_low_confidence": {"color": "#fec44f"},
            "deleterious": {"color": "#d53e4f"},
        }
        for c, cat in plot_cats.items():
            if c not in p_cats:
                p_cats[c] = cat
            p_cats[c].update(cat)
            p_cats[c]["name"] = p_cats[c]["name"].replace("_", " ").capitalize()

        plot_config["ylab"] = "Number of variants"

        self.add_section(
            name=title,
            anchor=htmlid,
            description="SIFT variant effect prediction.",
            plot=bargraph.plot(plot_data, p_cats, plot_config),
        )

    def bar_graph_polyphen(self):
        title = "PolyPhen summary"
        plot_data, plot_cats, plot_config = self._prep_bar_graph(title)
        htmlid = re.sub(r"\W+", "_", title).lower()
        if len(plot_data) == 0:
            return

        # Customise order and colours of categories
        p_cats = {
            "benign": {"color": "#a6db9f"},
            "possibly_damaging": {"color": "#fec44f"},
            "probably_damaging": {"color": "#d53e4f"},
            "unknown": {"color": "#d9d9d9"},
        }
        for c, cat in plot_cats.items():
            if c not in p_cats:
                p_cats[c] = cat
            p_cats[c].update(cat)
            p_cats[c]["name"] = p_cats[c]["name"].replace("_", " ").capitalize()

        plot_config["ylab"] = "Number of variants"

        self.add_section(
            name=title,
            anchor=htmlid,
            description="PolyPhen variant effect prediction.",
            plot=bargraph.plot(plot_data, p_cats, plot_config),
        )

    def bar_graph_variants_by_chromosome(self):
        title = "Variants by chromosome"
        plot_data, plot_cats, plot_config = self._prep_bar_graph(title)
        htmlid = re.sub(r"\W+", "_", title).lower()
        if len(plot_data) == 0:
            return

        # Sort the chromosomes numerically (almost - feel free to improve)
        chrs = {chr: k["name"].replace("chr", "").split("_")[0].rjust(20, "0") for chr, k in plot_cats.items()}
        p_cats = {}
        for chr in sorted(chrs, key=chrs.get):
            p_cats[chr] = plot_cats[chr]

        plot_config["cpswitch_c_active"] = False

        self.add_section(
            name=title,
            anchor=htmlid,
            description="Number of variants found on each chromosome.",
            plot=bargraph.plot(plot_data, p_cats, plot_config),
        )

    def bar_graph_position_in_protein(self):
        title = "Position in protein"
        plot_data, plot_cats, plot_config = self._prep_bar_graph(title)
        htmlid = re.sub(r"\W+", "_", title).lower()
        if len(plot_data) == 0:
            return

        # Nice graduated colours for categories
        colour_scale = mqc_colour.mqc_colour_scale("YlGnBu", 0, len(plot_cats) - 1)
        for idx, k in enumerate(plot_cats):
            plot_cats[k]["color"] = colour_scale.get_colour(idx, lighten=0.8)

        plot_config["cpswitch_c_active"] = False

        self.add_section(
            name=title,
            anchor=htmlid,
            description="Relative position of affected amino acids in protein.",
            plot=bargraph.plot(plot_data, plot_cats, plot_config),
        )

    def _prep_bar_graph(self, title):
        plot_data = dict()
        for s_name in self.vep_data:
            if title in self.vep_data[s_name]:
                plot_data[s_name] = self.vep_data[s_name][title]
        plot_cats = dict()
        htmlid = re.sub(r"\W+", "_", title).lower()
        plotid = f"{htmlid}_plot"
        plot_config = {
            "id": plotid,
            "title": f"VEP: {title}",
            "ylab": "Number of variants",
        }
        if len(plot_data) == 0:
            return [plot_data, plot_cats, plot_config]

        # Build plot categories
        for d in plot_data.values():
            for k in d.keys():
                if k not in plot_cats:
                    plot_cats[k] = {"name": k}

        return [plot_data, plot_cats, plot_config]



""" MultiQC module to parse output from DeDup """

import json
import logging
from json import JSONDecodeError

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph
from multiqc.utils import config

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """DeDup module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="DeDup",
            anchor="dedup",
            href="http://www.github.com/apeltzer/DeDup",
            info="is a tool for duplicate removal for merged/collapsed reads in ancient DNA analysis.",
            doi="10.1186/s13059-016-0918-z",
        )

        # Find and load any DeDup reports
        self.dedup_data = dict()

        for f in self.find_log_files("dedup", filehandles=True):
            try:
                self.parseJSON(f)
            except KeyError:
                logging.warning(f"Error loading file {f['fn']}")

        # Filter to strip out ignored sample names
        self.dedup_data = self.ignore_samples(self.dedup_data)

        if len(self.dedup_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.dedup_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.dedup_data, "multiqc_dedup")

        # Basic Stats Table
        self.dedup_general_stats_table()

        # Alignment Rate Plot
        self.add_section(
            description="This plot shows read categories that were either not removed (unique reads) or removed (duplicates).",
            plot=self.dedup_alignment_plot(),
        )

    # Parse our nice little JSON file
    def parseJSON(self, f):
        """Parse the JSON output from DeDup and save the summary statistics"""
        try:
            parsed_json = json.load(f["f"])
            # Check for Keys existing
            if "metrics" not in parsed_json or "metadata" not in parsed_json:
                log.debug(f"DeDup JSON missing essential keys - skipping sample: '{f['fn']}'")
                return None
        except JSONDecodeError as e:
            log.debug(f"Could not parse DeDup JSON: '{f['fn']}'")
            log.debug(e)
            return None

        # Get sample name from JSON first
        s_name = self.clean_s_name(parsed_json["metadata"]["sample_name"], f)
        self.add_data_source(f, s_name)

        # Add version info
        version = parsed_json["metadata"]["version"]
        self.add_software_version(version, s_name)

        metrics_dict = parsed_json["metrics"]

        for k in metrics_dict:
            metrics_dict[k] = float(metrics_dict[k])

        # Compute (not) removed _mapped_ reads from given values as dedup only affects mapped reads
        # Keep legacy behaviour in case "mapped_reads" cannot be found for <= v0.12.6
        if "mapped_reads" in metrics_dict:
            metrics_dict["mapped_after_dedup"] = (
                metrics_dict["mapped_reads"]
                - metrics_dict["reverse_removed"]
                - metrics_dict["forward_removed"]
                - metrics_dict["merged_removed"]
            )
        else:
            metrics_dict["not_removed"] = (
                metrics_dict["total_reads"]
                - metrics_dict["reverse_removed"]
                - metrics_dict["forward_removed"]
                - metrics_dict["merged_removed"]
            )
        metrics_dict["reads_removed"] = (
            metrics_dict["reverse_removed"] + metrics_dict["forward_removed"] + metrics_dict["merged_removed"]
        )

        # Add all in the main data_table
        self.dedup_data[s_name] = metrics_dict

    def dedup_general_stats_table(self):
        """Take the parsed stats from the DeDup report and add it to the
        basic stats table at the top of the report"""

        ancient_read_count_prefix = getattr(config, "ancient_read_count_prefix", "K")
        ancient_read_count_desc = getattr(config, "ancient_read_count_desc", "thousands")
        ancient_read_count_multiplier = getattr(config, "ancient_read_count_multiplier", 0.001)

        headers = {
            "dup_rate": {
                "title": "Duplication Rate",
                "description": "Percentage of reads categorised as a technical duplicate",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
                "format": "{:,.0f}",
                "modify": lambda x: x * 100.0,
            },
            "clusterfactor": {
                "title": "ClusterFactor",
                "description": "CF~1 means high library complexity. Large CF means not worth sequencing deeper.",
                "min": 1,
                "max": 100,
                "scale": "OrRd",
                "format": "{:,.2f}",
            },
            "reads_removed": {
                "title": f"{ancient_read_count_prefix} Reads Removed",
                "description": f"Non-unique reads removed after deduplication ({ancient_read_count_desc})",
                "modify": lambda x: x * ancient_read_count_multiplier,
                "shared_key": "read_count",
                "min": 0,
                "hidden": True,
            },
            "mapped_after_dedup": {
                "title": f"{ancient_read_count_prefix} Post-DeDup Mapped Reads",
                "description": f"Unique mapping reads after deduplication ({ancient_read_count_desc})",
                "modify": lambda x: x * ancient_read_count_multiplier,
                "shared_key": "read_count",
                "min": 0,
            },
        }
        self.general_stats_addcols(self.dedup_data, headers)

    def dedup_alignment_plot(self):
        """Make the HighCharts HTML to plot the duplication rates"""

        # Specify the order of the different possible categories
        keys = {
            "mapped_after_dedup": {"name": "Unique Retained"},
            "not_removed": {"name": "Not Removed"},
            "reverse_removed": {"name": "Reverse Removed"},
            "forward_removed": {"name": "Forward Removed"},
            "merged_removed": {"name": "Merged Removed"},
        }

        # Config for the plot
        config = {
            "id": "dedup_rates",
            "title": "DeDup: Deduplicated Reads",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        return bargraph.plot(self.dedup_data, keys, config)



def clean_title_case(col_id):
    title = col_id.title() if col_id[0:1].islower() else col_id
    for _str in ["Bc", "bc", "Umi", "Igk", "Igh", "Igl", "Vj", "q30"]:
        title = title.replace(_str, _str.upper())
    return title


def update_dict(table, headers, rows_list, col_map, colours, prefix, int_cols=()):
    """update the data dict and headers dict

    :param: int_cols columns to be shown as integers in the table
    """

    for col_name, col_data in rows_list:
        if col_name in col_map:
            # Sanitize numeric data
            is_percentage = "%" in col_data
            col_data = col_data.replace(",", "").replace("%", "")

            # Convert to float when possible
            if col_data == "None":
                col_data = None
            else:
                try:
                    col_data = float(col_data)
                except ValueError:
                    pass

            col_id = col_map[col_name]
            table[col_id] = col_data
            headers[col_id] = {
                "rid": "{}_{}".format(prefix, col_id.replace(" ", "_").replace("/", "_")),
                "title": clean_title_case(col_id),
                "description": col_name,
                "namespace": f"Space Ranger {prefix}",
                "scale": colours.get(col_id, "RdYlGn" if is_percentage else "GnBu"),
            }
            if is_percentage:
                headers[col_id].update(
                    {
                        "suffix": "%",
                        "max": 100,
                        "min": 0,
                    }
                )

            if col_id in int_cols:
                headers[col_id]["format"] = "{:,.0f}"


def set_hidden_cols(headers, col_names):
    """Set the hidden columns"""

    for col_name in col_names:
        try:
            headers[col_name]["hidden"] = True
        except KeyError:
            pass

    return headers


def parse_bcknee_data(data, s_name, max_idx=1000):
    """parse data for bc knee plot from spaceranger dict"""

    value_dict = dict()
    for idx, data_series in enumerate(data):
        if idx > max_idx:
            break
        if len(data_series["x"]) == 0:
            continue
        id = f"{s_name}_{data_series['name']}"
        if id not in value_dict.keys():
            value_dict[id] = dict()
        value_dict[id].update(transform_data(data_series))

    return value_dict


def transform_data(data):
    """Transform x:list,y:list data to a dict of x_val:y_val"""

    return {x: y for x, y in zip(data["x"], data["y"])}



""" MultiQC module to parse output from BUSCO """


import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"# BUSCO version is: ([\d\.]+)"


class MultiqcModule(BaseMultiqcModule):
    """BUSCO module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="BUSCO",
            anchor="busco",
            href="http://busco.ezlab.org/",
            info="assesses genome assembly and annotation completeness with Benchmarking Universal Single-Copy Orthologs.",
            doi="10.1093/bioinformatics/btv351",
        )

        # Keys and strings, used for parsing and for plot
        self.busco_keys = {
            "complete": "Complete BUSCOs",
            "complete_single_copy": "Complete and single-copy BUSCOs",
            "complete_duplicated": "Complete and duplicated BUSCOs",
            "fragmented": "Fragmented BUSCOs",
            "missing": "Missing BUSCOs",
            "total": "Total BUSCO groups searched",
        }

        # Find and load any BUSCO reports
        self.busco_data = dict()
        for f in self.find_log_files("busco"):
            self.parse_busco_log(f)

        # Filter to strip out ignored sample names
        self.busco_data = self.ignore_samples(self.busco_data)

        if len(self.busco_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.busco_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.busco_data, "multiqc_busco")

        # One Alignment Rate Plot per lineage
        lineages = set([self.busco_data[s_name].get("lineage_dataset") for s_name in self.busco_data.keys()])
        for lin in lineages:
            self.add_section(
                name="Lineage Assessment" if lin is None else f"Lineage: {lin}",
                anchor="busco-lineage-" + re.sub(r"\W+", "_", str(lin)),
                plot=self.busco_plot(lin),
            )

    def parse_busco_log(self, f):
        parsed_data = {}
        for line in f["f"].splitlines():
            if line.startswith("# BUSCO version is"):
                version_match = re.search(VERSION_REGEX, line)
                if version_match:
                    self.add_software_version(version_match.group(1), f["s_name"])

            for key, string in self.busco_keys.items():
                if string in line:
                    s = line.strip().split("\t")
                    parsed_data[key] = float(s[0])
            if "The lineage dataset is:" in line:
                s = line.replace("# The lineage dataset is: ", "").split(" (Creation date:", 1)
                parsed_data["lineage_dataset"] = str(s[0])

        if len(parsed_data) > 0:
            self.busco_data[f["s_name"]] = parsed_data
            self.add_data_source(f)

    def busco_plot(self, lin):
        """Make the HighCharts HTML for the BUSCO plot for a particular lineage"""

        data = {}
        for s_name in self.busco_data:
            if self.busco_data[s_name].get("lineage_dataset") == lin:
                data[s_name] = self.busco_data[s_name]

        plot_keys = ["complete_single_copy", "fragmented", "complete_duplicated", "missing"]
        plot_cols = ["#31a354", "#fee8c8", "#fdbb84", "#e34a33"]
        keys = dict()
        for k, col in zip(plot_keys, plot_cols):
            keys[k] = {"name": self.busco_keys[k], "color": col}

        # Config for the plot
        config = {
            "id": "busco_plot_" + re.sub(r"\W+", "_", str(lin)),
            "title": "BUSCO: Assessment Results" if lin is None else f"BUSCO Assessment Results: {lin}",
            "ylab": "# BUSCOs",
            "cpswitch_counts_label": "Number of BUSCOs",
        }

        return bargraph.plot(data, keys, config)



""" MultiQC module to parse output from methylQA """


import logging
import os
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="methylQA",
            anchor="methylqa",
            target="methylqa",
            href="http://methylqa.sourceforge.net/",
            info=" - a methylation sequencing data quality assessment tool.",
            doi="10.1016/j.ymeth.2014.10.032",
        )

        # Find and load any methylQA reports
        self.methylqa_data = dict()
        self.methylqa_coverage_counts = dict()
        self.methylqa_coverage_percentages = dict()
        for f in self.find_log_files("methylqa"):
            self.parse_methylqa_logs(f)

        # Filter to strip out ignored sample names
        self.methylqa_data = self.ignore_samples(self.methylqa_data)

        if len(self.methylqa_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.methylqa_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.methylqa_data, "multiqc_methylQA")

        # Basic Stats Table
        self.methylqa_general_stats_table()

        # Alignment bar plot - only one section, so add to the module intro
        self.add_section(plot=self.methylqa_alignment_plot())

    def parse_methylqa_logs(self, f):
        # Get s_name from first input file if possible
        s_name = f["s_name"]
        if f["f"][0].startswith("files provided"):
            s_name = self.clean_s_name(os.path.basename(f["f"][0]), f)

        parsed_data = {}
        regexes = {
            "mappable_reads": r"uniquely mappable reads \(pair\):\s*(\d+)",
            "quality_failed": r"quality failed mapped reads \(pair\) in the bismark bam:\s*(\d+)",
            "oversized_reads": r"oversized mapped reads \(pair\) in the bismark bam:\s*(\d+)",
            "mapped_bases": r"total base of uniquely mapped reads \(pair\):\s*(\d+)",
            "coverage": r"total base of uniquely mapped reads \(pair\) cover genome base \(\d+\):\s*([\d\.]+)X",
            "CHG_count_meth": r"number of methylated C in CHG context \(was protected\):\s*(\d+)",
            "CHG_count_unmeth": r"number of not methylated C in CHG context \(was converted\):\s*(\d+)",
            "CHG_percent_meth": r"C->T convertion rate in CHG context:\s*([\d\.]+)%",
            "CHH_count_meth": r"number of methylated C in CHH context \(was protected\):\s*(\d+)",
            "CHH_count_unmeth": r"number of not methylated C in CHH context \(was converted\):\s*(\d+)",
            "CHH_percent_meth": r"C->T convertion rate in CHH context:\s*([\d\.]+)%",
            "CG_count_meth": r"number of methylated C in CpG context \(was protected\):\s*(\d+)",
            "CG_count_unmeth": r"number of not methylated C in CpG context \(was converted\):\s*(\d+)",
            "CG_percent_meth": r"C->T convertion rate in CpG context:\s*([\d\.]+)%",
            "CN_count_meth": r"number of methylated C in Unknown context \(was protected\):\s*(\d+)",
            "CN_count_unmeth": r"number of not methylated C in Unknown context \(was converted\):\s*(\d+)",
            "CN_percent_meth": r"C->T convertion rate in Unknown context:\s*([\d\.]+)%",
        }
        for k, r in regexes.items():
            match = re.search(r, f["f"])
            if match:
                parsed_data[k] = float(match.group(1))

        # Parse the coverage histogram
        hist = False
        for line in f["f"].split(os.linesep):
            if hist is not False:
                try:
                    s = line.split()
                    hist["counts"][s[0]] = int(s[1])
                    hist["percentages"][s[0]] = float(s[2])
                except IndexError:
                    break
            if re.search(r"Times covered\s+Count\s+Percent\s+", line):
                hist = {"counts": dict(), "percentages": dict()}
        if hist is not False and len(hist) > 0:
            self.methylqa_coverage_counts[s_name] = hist["counts"]
            self.methylqa_coverage_percentages[s_name] = hist["percentages"]

        if len(parsed_data) > 0:
            if s_name in self.methylqa_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, s_name)
            self.methylqa_data[s_name] = parsed_data

    def methylqa_general_stats_table(self):
        """Take the parsed stats from the methylQA report and add it to the
        basic stats table at the top of the report"""

        headers = {"coverage": {"title": "Fold Coverage", "min": 0, "suffix": "X", "scale": "YlGn"}}
        self.general_stats_addcols(self.methylqa_data, headers)

    def methylqa_alignment_plot(self):
        """Make the HighCharts HTML to plot the alignment rates"""

        if len(self.methylqa_coverage_counts) == 0:
            return '<div class="alert alert-danger">No histogram data found.</div>'

        pconfig = {
            "id": "methylqa_coverage",
            "title": "methylQA: CpG Coverage",
            "categories": True,
            "ylab": "CpG Counts",
            "xlab": "Times covered",
            "ymin": 0,
            "data_labels": [
                {"name": "Counts", "ylab": "CpG Counts"},
                {"name": "Percentages", "ylab": "% CpGs", "ymax": 100},
            ],
        }
        return linegraph.plot([self.methylqa_coverage_counts, self.methylqa_coverage_percentages], pconfig)



""" MultiQC submodule to parse output from deepTools plotEnrichment """

import logging

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class PlotEnrichmentMixin:
    def parse_plot_enrichment(self):
        """Find plotEnrichment output."""
        self.deeptools_plotEnrichment = dict()
        for f in self.find_log_files("deeptools/plotEnrichment"):
            parsed_data = self.parsePlotEnrichment(f)
            for k, v in parsed_data.items():
                if k in self.deeptools_plotEnrichment:
                    log.warning(f"Replacing duplicate sample {k}.")
                self.deeptools_plotEnrichment[k] = v

            if len(parsed_data) > 0:
                self.add_data_source(f, section="plotEnrichment")

        self.deeptools_plotEnrichment = self.ignore_samples(self.deeptools_plotEnrichment)

        if len(self.deeptools_plotEnrichment) == 0:
            return 0

        # Write data to file
        self.write_data_file(self.deeptools_plotEnrichment, "deeptools_plot_enrich")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        dCounts = dict()
        dPercents = dict()
        for sample, v in self.deeptools_plotEnrichment.items():
            dCounts[sample] = dict()
            dPercents[sample] = dict()
            for category, v2 in v.items():
                dCounts[sample][category] = v2["count"]
                dPercents[sample][category] = v2["percent"]
        config = {
            "data_labels": [
                {"name": "Counts in features", "ylab": "Counts in feature"},
                {"name": "Percents in features", "ylab": "Percent of reads in feature"},
            ],
            "id": "deeptools_enrichment_plot",
            "title": "deepTools: Signal enrichment per feature",
            "ylab": "Counts in feature",
            "categories": True,
            "ymin": 0.0,
        }
        self.add_section(
            name="Feature enrichment",
            description="Signal enrichment per feature according to plotEnrichment",
            anchor="deeptools_enrichment",
            plot=linegraph.plot([dCounts, dPercents], pconfig=config),
        )

        return len(self.deeptools_plotEnrichment)

    def parsePlotEnrichment(self, f):
        d = {}
        firstLine = True
        for line in f["f"].splitlines():
            if firstLine:
                firstLine = False
                continue
            cols = line.strip().split("\t")

            if len(cols) != 5:
                log.warning(
                    "{} was initially flagged as the output from plotEnrichment, but that seems to not be the case. Skipping...".format(
                        f["fn"]
                    )
                )
                return dict()

            s_name = self.clean_s_name(cols[0], f)
            if s_name not in d:
                d[s_name] = dict()
            cols[1] = str(cols[1])
            if cols[1] in d[s_name]:
                log.warning(f"Replacing duplicate sample:featureType {s_name}:{cols[1]}.")
            d[s_name][cols[1]] = dict()

            try:
                d[s_name][cols[1]]["percent"] = float(cols[2])
                d[s_name][cols[1]]["count"] = self._int(cols[3])
            except:  # noqa: E722
                log.warning(
                    "{} was initially flagged as the output from plotEnrichment, but that seems to not be the case. Skipping...".format(
                        f["fn"]
                    )
                )
                return dict()
        return d



""" MultiQC module to parse output from fgbio """
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

from . import ErrorRateByReadPosition
from .groupreadsbyumi import GroupReadsByUmiMixin

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule, GroupReadsByUmiMixin):
    """fgbio has a number of different commands and outputs.
    This MultiQC module supports some but not all. The code for
    each script is split into its own file and adds a section to
    the module output if logs are found."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="fgbio",
            anchor="fgbio",
            target="fgbio",
            href="http://fulcrumgenomics.github.io/fgbio/",
            info=" is a command line toolkit for working with genomic and particularly next generation sequencing data..",
            # No publication / DOI // doi=
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = dict()
        self.general_stats_data = dict()

        # GroupReadsByUmi
        n = dict()
        n["groupreadsbyumi"] = self.parse_groupreadsbyumi()
        if n["groupreadsbyumi"] > 0:
            log.info(f"Found {n['groupreadsbyumi']} groupreadsbyumi reports")

        # ErrorRateByReadPoosition
        n["errorratebyreadposition"] = ErrorRateByReadPosition.parse_reports(self)
        if n["errorratebyreadposition"] > 0:
            log.info(f"Found {n['errorratebyreadposition']} errorratebyreadposition reports")

        # Exit if we didn't find anything
        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound

        # Add to the General Stats table (has to be called once per MultiQC module)
        self.general_stats_addcols(self.general_stats_data, self.general_stats_headers)



""" MultiQC module to parse output from biobambam2 """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.modules.picard import MarkDuplicates

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """This module is super weird. The output from this tools is essentially
    identical to Picard MarkDuplicates, so we just hijack that module instead"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="biobambam2",
            anchor="biobambam2",
            href="https://gitlab.com/german.tischler/biobambam2",
            info="provides tools for early stage alignment file processing",
            doi="10.1186/1751-0473-9-13",
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = dict()
        self.general_stats_data = dict()
        n = dict()

        n["bamsormadup"] = MarkDuplicates.parse_reports(self, "biobambam2/bamsormadup")
        if n["bamsormadup"] > 0:
            log.info(f"Found {n['bamsormadup']} bamsormadup reports")

        # Exit if we didn't find anything
        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound

        # Add to the General Stats table (has to be called once per MultiQC module)
        self.general_stats_addcols(self.general_stats_data, self.general_stats_headers)

    # Helper functions
    @staticmethod
    def multiply_hundred(val):
        try:
            val = float(val) * 100
        except ValueError:
            pass
        return val



import logging
import os

from ..base_module import ModuleNoSamplesFound
from .base_metrics import DragenBaseMetrics
from .content_metrics import DragenContentMetrics
from .gc_metrics import DragenFastqcGcMetrics
from .read_metrics import DragenReadMetrics
from .util import parse_fastqc_metrics_file

log = logging.getLogger(__name__)


class MultiqcModule(DragenBaseMetrics, DragenReadMetrics, DragenFastqcGcMetrics, DragenContentMetrics):
    """DRAGEN provides a number of different pipelines and outputs, including base calling, DNA and RNA alignment,
    post-alignment processing and variant calling, covering virtually all stages of typical NGS data processing.
    However, it can be treated as a fast aligner with additional features on top, as users will unlikely use any
    features without enabling DRAGEN mapping. So we will treat this module as an alignment tool module and
    place it accordingly in the module_order list, in docs, etc.

    The QC metrics DRAGEN generates resemble those of samtools-stats, qualimap, mosdepth, bcftools-stats and alike.
    Whenver possible, the visual output is made similar to those modules.

    Note that this MultiQC module supports some of DRAGEN output but not all. Contributions are welcome!

    The code is structured in a way so every mix-in parses one type of QC file that DRAGEN generates
    (e.g. *.mapping_metrics.csv, *.wgs_fine_hist_normal.csv, etc). If a corresponding file is found, a mix-in adds
    a section into the report.
    """

    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="DRAGEN-FastQC",
            anchor="dragen-fastqc",
            href="https://www.illumina.com/products/by-type/informatics-products/dragen-bio-it-platform.html",
            info=(
                " is a Bio-IT Platform that provides ultra-rapid secondary analysis of sequencing data"
                " using field-programmable gate array technology (FPGA)."
            ),
            # Can't find a DOI // doi=
        )

        self.css = {
            "assets/css/multiqc_fastqc.css": os.path.join(
                os.path.dirname(__file__), "..", "fastqc", "assets", "css", "multiqc_fastqc.css"
            )
        }
        self.js = {
            "assets/js/multiqc_dragen_fastqc.js": os.path.join(
                os.path.dirname(__file__), "assets", "js", "multiqc_dragen_fastqc.js"
            )
        }
        self.intro += '<script type="application/json" class="fastqc_passfails">["dragen_fastqc", {"per_base_sequence_content": {"TEST": "pass"}}]</script>'

        data_by_sample = {}
        for f in self.find_log_files("dragen_fastqc"):
            s_name, data_by_mate = parse_fastqc_metrics_file(f)

            # Clean sample name and key
            new_s_name = self.clean_s_name(s_name, f)
            data_by_mate[new_s_name] = data_by_mate.pop(s_name)
            s_name = new_s_name
            del new_s_name

            if s_name in data_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f, s_name)
            data_by_sample.update(data_by_mate)

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, s_name)

        # Filter to strip out ignored sample names:
        self.dragen_fastqc_data = self.ignore_samples(data_by_sample)

        # TODO: Split this up and write the interesting bits to files
        # Currently is 13760 lines in my test, which is too much.
        # self.write_data_file(self.dragen_fastqc_data, "dragen_fastqc")

        samples_found = set()

        # "POSITIONAL QUALITY" and "POSITIONAL BASE MEAN QUALITY" metrics
        samples_found |= self.add_base_metrics()

        # "READ MEAN QUALITY" and "READ LENGTHS" metrics
        samples_found |= self.add_read_metrics()

        # "GC CONTENT" and "GC CONTENT QUALITY" metrics
        samples_found |= self.add_gc_metrics()

        # "POSITIONAL BASE CONTENT" metrics
        samples_found |= self.add_content_metrics()

        if len(samples_found) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(samples_found)} reports")



""" MultiQC submodule to parse output from Picard ValidateSamFile """

import logging

from multiqc.plots import table

# Initialise the logger
log = logging.getLogger(__name__)

# Possible warnings and descriptions
WARNING_DESCRIPTIONS = {
    "ADJACENT_INDEL_IN_CIGAR": "CIGAR string contains an insertion (I) followed by " "deletion (D), or vice versa",
    "BAM_FILE_MISSING_TERMINATOR_BLOCK": "BAM appears to be healthy, but is an older "
    "file so doesn't have terminator block",
    "E2_BASE_EQUALS_PRIMARY_BASE": "Secondary base calls should not be the same as "
    "primary, unless one or the other is N",
    "INVALID_DATE_STRING": "Date string is not ISO-8601",
    "INVALID_QUALITY_FORMAT": "Quality encodings out of range; appear to be Solexa or "
    "Illumina when Phred expected. Avoid exception being "
    "thrown as a result of no qualities being read.",
    "MISSING_TAG_NM": "The NM tag (nucleotide differences) is missing",
    "PAIRED_READ_NOT_MARKED_AS_FIRST_OR_SECOND": "Pair flag set but not marked as " "first or second of pair",
    "RECORD_MISSING_READ_GROUP": "A SAMRecord is found with no read group id",
}

# All possible errors and descriptions
ERROR_DESCRIPTIONS = {
    "CIGAR_MAPS_OFF_REFERENCE": "Bases corresponding to M operator in CIGAR extend " "beyond reference",
    "DUPLICATE_PROGRAM_GROUP_ID": "Same program group id appears more than once",
    "DUPLICATE_READ_GROUP_ID": "Same read group id appears more than once",
    "EMPTY_READ": "Indicates that a read corresponding to the first strand has a "
    "length of zero and/or lacks flow signal intensities (FZ)",
    "HEADER_RECORD_MISSING_REQUIRED_TAG": "Header tag missing in header line",
    "HEADER_TAG_MULTIPLY_DEFINED": "Header tag appears more than once in header line " "" "" "with different value",
    "INVALID_ALIGNMENT_START": "Alignment start position is incorrect",
    "INVALID_CIGAR": "CIGAR string error for either read or mate",
    "INVALID_FLAG_FIRST_OF_PAIR": "First of pair flag set for unpaired read",
    "INVALID_FLAG_MATE_NEG_STRAND": "Mate negative strand flag set for unpaired read",
    "INVALID_FLAG_MATE_UNMAPPED": "Mate unmapped flag is incorrectly set",
    "INVALID_FLAG_NOT_PRIM_ALIGNMENT": "Not primary alignment flag set for unmapped " "read",
    "INVALID_FLAG_PROPER_PAIR": "Proper pair flag set for unpaired read",
    "INVALID_FLAG_READ_UNMAPPED": "Mapped read flat not set for mapped read",
    "INVALID_FLAG_SECOND_OF_PAIR": "Second of pair flag set for unpaired read",
    "INVALID_FLAG_SUPPLEMENTARY_ALIGNMENT": "Supplementary alignment flag set for " "unmapped read",
    "INVALID_INDEX_FILE_POINTER": "Invalid virtualFilePointer in index",
    "INVALID_INDEXING_BIN": "Indexing bin set on SAMRecord does not agree with " "computed value",
    "INVALID_INSERT_SIZE": "Inferred insert size is out of range",
    "INVALID_MAPPING_QUALITY": "Mapping quality set for unmapped read or is >= 256",
    "INVALID_MATE_REF_INDEX": "Mate reference index (MRNM) set for unpaired read",
    "INVALID_PLATFORM_VALUE": "The read group has an invalid value set for its PL " "field",
    "INVALID_PREDICTED_MEDIAN_INSERT_SIZE": "PI tag value is not numeric",
    "INVALID_REFERENCE_INDEX": "Reference index not found in sequence dictionary",
    "INVALID_TAG_NM": "The NM tag (nucleotide differences) is incorrect",
    "INVALID_VERSION_NUMBER": "Does not match any of the acceptable versions",
    "MATE_CIGAR_STRING_INVALID_PRESENCE": "A cigar string for a read whose mate is " "NOT mapped",
    "MATE_FIELD_MISMATCH": "Read alignment fields do not match its mate",
    "MATE_NOT_FOUND": "Read is marked as paired, but its pair was not found",
    "MATES_ARE_SAME_END": "Both mates of a pair are marked either as first or second " "" "" "mates",
    "MISMATCH_FLAG_MATE_NEG_STRAND": "Mate negative strand flag does not match read " "strand flag",
    "MISMATCH_FLAG_MATE_UNMAPPED": "Mate unmapped flag does not match read unmapped " "flag of mate",
    "MISMATCH_MATE_ALIGNMENT_START": "Mate alignment does not match alignment start " "of mate",
    "MISMATCH_MATE_CIGAR_STRING": "The mate cigar tag does not match its mate's cigar " "" "" "string",
    "MISMATCH_MATE_REF_INDEX": "Mate reference index (MRNM) does not match reference " "" "" "index of mate",
    "MISMATCH_READ_LENGTH_AND_E2_LENGTH": "Lengths of secondary base calls tag " "values" " and read should match",
    "MISMATCH_READ_LENGTH_AND_QUALS_LENGTH": "Length of sequence string and length of"
    " base quality string do not match",
    "MISMATCH_READ_LENGTH_AND_U2_LENGTH": "Secondary base quals tag values should " "match read length",
    "MISSING_HEADER": "The SAM/BAM file is missing the header",
    "MISSING_PLATFORM_VALUE": "The read group is missing its PL (platform unit) field",
    "MISSING_READ_GROUP": "The header is missing read group information",
    "MISSING_SEQUENCE_DICTIONARY": "There is no sequence dictionary in the header",
    "MISSING_VERSION_NUMBER": "Header has no version number",
    "POORLY_FORMATTED_HEADER_TAG": "Header tag does not have colon",
    "READ_GROUP_NOT_FOUND": "A read group ID on a SAMRecord is not found in the header",
    "RECORD_OUT_OF_ORDER": "The record is out of order",
    "TAG_VALUE_TOO_LARGE": "Unsigned integer tag value is deprecated in BAM. Template " "" "" "length",
    "TRUNCATED_FILE": "BAM file does not have terminator block",
    "UNRECOGNIZED_HEADER_TYPE": "Header record is not one of the standard types",
}


def _default_data_entry():
    return {"WARNING_count": 0, "ERROR_count": 0, "file_validation_status": "pass"}


def parse_reports(module):
    """
    Find Picard ValidateSamFile reports and parse their data based on wether we
    think it's a VERBOSE or SUMMARY report
    """

    # Get data
    data = _parse_reports_by_type(module)

    if data:
        #  Filter to strip out ignored sample names (REQUIRED)
        data = module.ignore_samples(data)

        # Populate the general stats table
        module.general_stats_addcols(data, _get_general_stats_headers(), namespace="ValidateSamFile")

        # Add any found data to the report
        _add_section_to_report(module, data)

        # Write parsed data to a file
        module.write_data_file(data, "multiqc_picard_validatesamfile")

    return len(data)


def _parse_reports_by_type(module):
    """Returns a data dictionary

    Goes through logs and parses them based on 'No errors found', VERBOSE or SUMMARY
    type.
    """

    data_by_sample = dict()

    for f in module.find_log_files("picard/sam_file_validation", filehandles=True):
        module.add_data_source(f, "ValidateSamFile")

        s_name = f["s_name"]

        if s_name in data_by_sample:
            log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")

        fh = f["f"]
        first_line = fh.readline().rstrip()
        fh.seek(0)  # Rewind reading of the file

        if "No errors found" in first_line:
            sample_data = _parse_no_error_report()
        elif first_line.startswith("ERROR") or first_line.startswith("WARNING"):
            sample_data = _parse_verbose_report(fh)
        else:
            sample_data = _parse_summary_report(fh)

        data_by_sample[s_name] = sample_data

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        module.add_software_version(None, s_name)

    return data_by_sample


def _parse_no_error_report():
    return _default_data_entry()


def _parse_verbose_report(filehandle):
    sample_data = _default_data_entry()
    errors = warnings = 0
    for line in filehandle:
        if line.startswith("WARNING:"):
            warnings += 1
        elif line.startswith("ERROR:"):
            errors += 1

    sample_data["WARNING_count"] = warnings
    sample_data["ERROR_count"] = errors

    if errors:
        sample_data["file_validation_status"] = "fail"
    elif warnings:
        sample_data["file_validation_status"] = "warn"

    return sample_data


def _parse_summary_report(filehandle):
    sample_data = _default_data_entry()
    for problem_type, name, count in _histogram_data(filehandle):
        sample_data[name] = count
        sample_data[problem_type + "_count"] += count

    if sample_data["ERROR_count"]:
        sample_data["file_validation_status"] = "fail"
    elif sample_data["WARNING_count"]:
        sample_data["file_validation_status"] = "warn"

    return sample_data


def _histogram_data(iterator):
    """Yields only the row contents that contain the histogram entries"""
    histogram_started = False
    header_passed = False
    for line in iterator:
        if "## HISTOGRAM" in line:
            histogram_started = True
        elif histogram_started:
            if header_passed:
                values = line.rstrip().split("\t")
                if len(values) == 1:
                    continue
                try:
                    problem_type, name = values[0].split(":")
                except ValueError:
                    log.warning(
                        "Line did not look like normal picard 'ERROR:NAME' format, " "" "" "ignoring: {}".format(
                            values[0]
                        )
                    )
                    continue
                yield problem_type, name, int(values[1])
            elif line.startswith("Error Type"):
                header_passed = True


def _add_section_to_report(module, data):
    """
    Adds found data to the report via several HTML generators
    """

    # Count samples that have errors and/or warnings
    pass_count = error_count = only_warning_count = 0
    for sample_data in data.values():
        if sample_data["file_validation_status"] == "fail":
            error_count += 1
        elif sample_data["file_validation_status"] == "warn":
            only_warning_count += 1
        else:
            pass_count += 1

    # Add overview note
    plot_html = []
    note_html = _generate_overview_note(
        pass_count=pass_count, only_warning_count=only_warning_count, error_count=error_count, total_count=len(data)
    )
    plot_html.append(note_html)

    # Add the detailed table, but only if we have something to show
    if error_count or only_warning_count:
        table_html = _generate_detailed_table(data)
        plot_html.append(table_html)

    # Finally, add the html to the report as a section
    module.add_section(
        name="SAM/BAM File Validation",
        anchor="picard_validatesamfile",
        description=(
            "This tool reports on the validity of a SAM or BAM " "file relative to " "the SAM-format specification."
        ),
        helptext="""
            A detailed table is only shown if errors or warnings are found. Details 
            about the errors and warnings are only shown if a `SUMMARY` report was 
            parsed.

            For more information on the warnings, errors and possible fixes please 
            read [this broadinstitute article](
            https://software.broadinstitute.org/gatk/documentation/article.php?id
            =7571).""",
        plot="\n".join(plot_html),
    )


def _get_general_stats_headers():
    """
    Returns a header dict for the general stats columns
    """
    headers = {
        "file_validation_status": {
            "title": "Validation",
            "description": "ValidateSamFile (File Validation)",
        },
        "WARNING_count": {
            "title": "# Warnings",
            "description": "ValidateSamFile (number of warnings)",
            "scale": "Oranges",
            "shared_key": "ValidateSamEntries",
            "colour": "255,237,160",
            "format": "{:.0f}",
            "hidden": True,
        },
        "ERROR_count": {
            "title": "# Errors",
            "description": "ValidateSamFile (number of errors)",
            "scale": "Reds",
            "shared_key": "ValidateSamEntries",
            "colour": "252,146,114",
            "format": "{:.0f}",
            "hidden": True,
        },
    }

    return headers


def _generate_overview_note(pass_count, only_warning_count, error_count, total_count):
    """Generates and returns the HTML note that provides a summary of validation
    status."""

    note_html = ['<div class="progress">']
    pbars = [
        [float(error_count), "danger", "had errors"],
        [float(only_warning_count), "warning", "had warnings"],
        [float(pass_count), "success", "passed"],
    ]
    for b in pbars:
        if b[0]:
            note_html.append(
                '<div class="progress-bar progress-bar-{pbcol}" style="width: {pct}%" '
                'data-toggle="tooltip" title="{count} {sample} {txt}">{'
                "count}</div>".format(
                    pbcol=b[1],
                    count=int(b[0]),
                    pct=(b[0] / float(total_count)) * 100.0,
                    txt=b[2],
                    sample="samples" if b[0] > 1 else "sample",
                )
            )
    note_html.append("</div>")

    return "\n".join(note_html)


def _generate_detailed_table(data):
    """
    Generates and returns the HTML table that overviews the details found.
    """
    headers = _get_general_stats_headers()

    # Only add headers for errors/warnings we have found
    for problems in data.values():
        for problem in problems:
            if problem not in headers and problem in WARNING_DESCRIPTIONS:
                headers["WARNING_count"]["hidden"] = False
                headers[problem] = {
                    "description": WARNING_DESCRIPTIONS[problem],
                    "namespace": "WARNING",
                    "scale": headers["WARNING_count"]["scale"],
                    "format": "{:.0f}",
                    "shared_key": "warnings",
                    "hidden": True,  # Hide by default; to unclutter things.
                }
            if problem not in headers and problem in ERROR_DESCRIPTIONS:
                headers["ERROR_count"]["hidden"] = False
                headers[problem] = {
                    "description": ERROR_DESCRIPTIONS[problem],
                    "namespace": "ERROR",
                    "scale": headers["ERROR_count"]["scale"],
                    "format": "{:.0f}",
                    "shared_key": "errors",
                    "hidden": True,  # Hide by default; to unclutter things.
                }

    table_config = {
        "table_title": "Picard: SAM/BAM File Validation",
        "id": "picard_validatesamfile_table",
    }

    return table.plot(data=data, headers=headers, pconfig=table_config)



""" MultiQC submodule to parse output from RSeQC tin.py
http://rseqc.sourceforge.net/#tin-py """

import csv
import logging


# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find RSeQC tin reports and parse their data"""

    self.tin_data = dict()

    for f in self.find_log_files("rseqc/tin", filehandles=True):
        # Parse contents
        try:
            reader = csv.DictReader(f["f"], delimiter="\t")
            contents = next(reader)
        except (csv.Error, StopIteration) as e:
            log.error(f"Could not parse file '{f['fn']}': {e}")
            continue

        s_name = self.clean_s_name(contents["Bam_file"], f)
        contents.pop("Bam_file")
        self.tin_data[s_name] = contents

        # Add file to multiqc_sources.txt
        self.add_data_source(f)

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None, s_name)

    # Filter to strip out ignored sample names
    self.tin_data = self.ignore_samples(self.tin_data)

    if len(self.tin_data) > 0:
        # Write to file
        self.write_data_file(self.tin_data, "multiqc_rseqc_tin")

        # Add to general stats table
        self.general_stats_headers["TIN(stdev)"] = {
            "title": "TIN stdev",
            "description": "Standard Deviation for the Transcript Integriry Number (TIN)",
            "max": 100,
            "min": 0,
            "scale": "Reds",
            "hidden": True,
        }
        self.general_stats_headers["TIN(median)"] = {
            "title": "TIN",
            "description": "Median Transcript Integriry Number (TIN), indicating the RNA integrity of a sample",
            "max": 100,
            "min": 0,
            "scale": "RdBu",
        }

        for s_name in self.tin_data:
            if s_name not in self.general_stats_data:
                self.general_stats_data[s_name] = dict()
            self.general_stats_data[s_name]["TIN(median)"] = self.tin_data[s_name]["TIN(median)"]
            self.general_stats_data[s_name]["TIN(stdev)"] = self.tin_data[s_name]["TIN(stdev)"]

    return len(self.tin_data)



# coding: utf-8
""" MultiQC submodule to parse output from Samtools flagstat """

import logging
import re

from multiqc import config
from multiqc.plots import beeswarm

# Initialise the logger
log = logging.getLogger(__name__)


class FlagstatReportMixin:
    def parse_samtools_flagstats(self):
        """Find Samtools flagstat logs and parse their data"""

        self.samtools_flagstat = dict()
        for f in self.find_log_files("samtools/flagstat"):
            parsed_data = parse_single_report(f["f"])
            if len(parsed_data) > 0:
                if f["s_name"] in self.samtools_flagstat:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                self.add_data_source(f, section="flagstat")
                self.samtools_flagstat[f["s_name"]] = parsed_data

        # Filter to strip out ignored sample names
        self.samtools_flagstat = self.ignore_samples(self.samtools_flagstat)

        if len(self.samtools_flagstat) == 0:
            return 0

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file (restructure first)
        self.write_data_file(self.samtools_flagstat, "multiqc_samtools_flagstat")

        # General Stats Table
        flagstats_headers = {
            "flagstat_total": {
                "title": "Reads",
                "description": f"Total reads in the bam file ({config.read_count_desc})",
                "shared_key": "read_count",
                "hidden": True,
            },
            "mapped_passed": {
                "title": "Reads mapped",
                "description": f"Reads mapped in the bam file ({config.read_count_desc})",
                "shared_key": "read_count",
            },
            "mapped_passed_pct": {
                "title": "% Reads mapped",
                "description": "% Reads mapped in the bam file",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "RdYlGn",
                "hidden": True,
            },
        }
        self.general_stats_addcols(self.samtools_flagstat, flagstats_headers, namespace="flagstat")

        # Make dot plot of counts
        keys = {}
        reads = {
            "min": 0,
            "modify": lambda x: float(x) * config.read_count_multiplier,
            "suffix": config.read_count_prefix,
            "decimalPlaces": 2,
            "shared_key": "read_count",
        }
        keys["flagstat_total"] = dict(reads, title="Total Reads")
        keys["total_passed"] = dict(reads, title="Total Passed QC")
        keys["mapped_passed"] = dict(reads, title="Mapped")

        if any(v.get("secondary_passed") for v in self.samtools_flagstat.values()):
            keys["secondary_passed"] = dict(reads, title="Secondary Alignments")

        if any(v.get("supplementary_passed") for v in self.samtools_flagstat.values()):
            keys["supplementary_passed"] = dict(reads, title="Supplementary Alignments")

        keys["duplicates_passed"] = dict(reads, title="Duplicates")
        keys["paired in sequencing_passed"] = dict(reads, title="Paired in Sequencing")
        keys["properly paired_passed"] = dict(reads, title="Properly Paired")
        keys["with itself and mate mapped_passed"] = dict(
            reads, title="Self and mate mapped", description="Reads with itself and mate mapped"
        )
        keys["singletons_passed"] = dict(reads, title="Singletons")
        keys["with mate mapped to a different chr_passed"] = dict(
            reads, title="Mate mapped to diff chr", description="Mate mapped to different chromosome"
        )
        keys["with mate mapped to a different chr (mapQ >= 5)_passed"] = dict(
            reads, title="Diff chr (mapQ >= 5)", description="Mate mapped to different chromosome (mapQ >= 5)"
        )

        self.add_section(
            name="Flagstat",
            anchor="samtools-flagstat",
            description="This module parses the output from <code>samtools flagstat</code>. All numbers in millions.",
            plot=beeswarm.plot(
                self.samtools_flagstat,
                keys,
                {
                    "id": "samtools-flagstat-dp",
                    "title": "Samtools flagstat: Read Counts",
                },
            ),
        )

        # Return the number of logs that were found
        return len(self.samtools_flagstat)


#  flagstat has one thing per line, documented here (search for flagstat):
#  http://www.htslib.org/doc/samtools.html
flagstat_regexes = {
    "total": r"(\d+) \+ (\d+) in total \(QC-passed reads \+ QC-failed reads\)",
    "secondary": r"(\d+) \+ (\d+) secondary",
    "supplementary": r"(\d+) \+ (\d+) supplementary",
    "duplicates": r"(\d+) \+ (\d+) duplicates",
    "mapped": r"(\d+) \+ (\d+) mapped \((.+):(.+)\)",
    "paired in sequencing": r"(\d+) \+ (\d+) paired in sequencing",
    "read1": r"(\d+) \+ (\d+) read1",
    "read2": r"(\d+) \+ (\d+) read2",
    "properly paired": r"(\d+) \+ (\d+) properly paired \((.+):(.+)\)",
    "with itself and mate mapped": r"(\d+) \+ (\d+) with itself and mate mapped",
    "singletons": r"(\d+) \+ (\d+) singletons \((.+):(.+)\)",
    "with mate mapped to a different chr": r"(\d+) \+ (\d+) with mate mapped to a different chr",
    "with mate mapped to a different chr (mapQ >= 5)": r"(\d+) \+ (\d+) with mate mapped to a different chr \(mapQ>=5\)",
}


def parse_single_report(file_obj):
    """
    Take a filename, parse the data assuming it's a flagstat file
    Returns a dictionary {'lineName_pass' : value, 'lineName_fail' : value}
    """
    parsed_data = {}

    re_groups = ["passed", "failed", "passed_pct", "failed_pct"]
    for k, r in flagstat_regexes.items():
        r_search = re.search(r, file_obj, re.MULTILINE)
        if r_search:
            for i, j in enumerate(re_groups):
                try:
                    key = f"{k}_{j}"
                    val = r_search.group(i + 1).strip("% ")
                    parsed_data[key] = float(val) if ("." in val) else int(val)
                except IndexError:
                    pass  # Not all regexes have percentages
                except ValueError:
                    parsed_data[key] = float("nan")
    # Work out the total read count
    try:
        parsed_data["flagstat_total"] = parsed_data["total_passed"] + parsed_data["total_failed"]
    except KeyError:
        pass
    return parsed_data



""" MultiQC module to parse output from SeqWho """

import json
import logging
from collections import OrderedDict

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph, bargraph
from multiqc.utils import mqc_colour

# Initialise the logger
log = logging.getLogger(__name__)


LIBRARIES = ["amplicon", "bisulf", "wxs", "chip", "wgs", "dnase", "rnaseq", "atac", "mirnaseq"]
SPECIES = ["human", "mouse"]


class MultiqcModule(BaseMultiqcModule):
    """SeqWho module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="SeqWho",
            anchor="seqwho",
            href="https://daehwankimlab.github.io/seqwho/",
            info="is a tool to determine a FASTQ(A) sequencing file identity, both source protocol and species of origin.",
            # doi=""
        )

        # Find and load any SeqWho reports
        self.seqwho_data = dict()
        self.seqwho_qualdis = dict()
        self.seqwho_qualscore = dict()
        self.seqwho_readdist = dict()

        for f in self.find_log_files("seqwho", filehandles=True):
            self.parse_json(f)

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Filter to strip out ignored sample names
        self.seqwho_data = self.ignore_samples(self.seqwho_data)
        self.seqwho_qualdis = self.ignore_samples(self.seqwho_qualdis)
        self.seqwho_qualscore = self.ignore_samples(self.seqwho_qualscore)
        self.seqwho_readdist = self.ignore_samples(self.seqwho_readdist)

        if len(self.seqwho_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.seqwho_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.seqwho_data, "multiqc_seqwho")

        # General Stats Table
        self.seqwho_general_stats_table()

        # SeqWho MLE Species Table
        self.seqwho_species_plot()

        # SeqWho MLE Sequencing Table
        self.seqwho_library_plot()

        # Quality Distribution
        self.seqwho_qualdist_plot()

        # Quality Score
        self.seqwho_qualscore_plot()

        # Read Distribution
        self.seqwho_readdist_plot()

    # Parse our nice little JSON file
    def parse_json(self, f):
        """Parse the JSON output from SeqWho and save the summary statistics"""
        try:
            parsed_json = json.load(f["f"])
        except json.JSONDecodeError:
            log.warning(f"Could not parse SeqWho JSON: '{f['fn']}'")
            return

        # Get sample name from JSON first
        for s_name, data in parsed_json.items():
            s_name = self.clean_s_name(s_name, f["root"])
            self.add_data_source(f, s_name)
            self.seqwho_data[s_name] = {}

            call_data = data["Call"]
            self.seqwho_data[s_name]["predicted_species"] = call_data["Species"]
            self.seqwho_data[s_name]["predicted_library"] = call_data["Library"]
            self.seqwho_data[s_name]["mle"] = call_data["Maximum Likelihood Estimate"]
            self.seqwho_data[s_name]["est_read_number"] = data["Estimated Read Number"]

            # Make MLE Matrix table for subset of predicted species and library
            mle_matrix = call_data["MLE Matrix"]
            mle_columns = SPECIES + LIBRARIES
            for i in mle_columns:
                self.seqwho_data[s_name][i] = mle_matrix[i]

            # Make a Quality Score vs count  version of the data
            self.seqwho_qualdis[s_name] = OrderedDict()
            qual_dist = data["Quality Dist"]
            for i in range(len(qual_dist)):
                self.seqwho_qualdis[s_name][i] = qual_dist[i]

            # Make a Quality score vs position  version of the data
            self.seqwho_qualscore[s_name] = OrderedDict()
            qual_scores = data["Quality Scores"]
            for i in range(len(qual_scores)):
                self.seqwho_qualscore[s_name][i] = qual_scores[i]

            # Make a Read Length vs count of the data
            self.seqwho_readdist[s_name] = OrderedDict()
            read_dist = data["Read lengths"]
            for i in range(len(read_dist)):
                self.seqwho_readdist[s_name][i] = read_dist[i]

    def seqwho_general_stats_table(self):
        """Take the parsed stats from the SeqWho report and add it to the
        General Statistics table at the top of the report"""
        species_scale = mqc_colour.mqc_colour_scale("Dark2")
        species_colors = [{v: species_scale.get_colour(i, lighten=0.5)} for i, v in enumerate(SPECIES)]
        library_scale = mqc_colour.mqc_colour_scale("Accent")
        library_colors = [{v: library_scale.get_colour(i, lighten=0.5)} for i, v in enumerate(LIBRARIES)]

        headers = {
            "predicted_species": {
                "title": "Pred Species",
                "description": "Predicted species",
                "cond_formatting_colours": species_colors,
                "cond_formatting_rules": {v: [{"s_eq": v}] for v in SPECIES},
                "scale": False,
            },
            "predicted_library": {
                "title": "Pred Library",
                "description": "Predicted library type",
                "cond_formatting_colours": library_colors,
                "cond_formatting_rules": {v: [{"s_eq": v}] for v in LIBRARIES},
                "scale": False,
            },
            "mle": {
                "title": "Max Likelihood Est",
                "description": "Overall maximum likelihood estimate (overall)",
                "min": 0,
                "format": "{:,.3f}",
                "scale": "RdYlGn",
            },
            "est_read_number": {
                "title": f"{config.read_count_prefix} Est Reads",
                "description": f"Estimated read number ({config.read_count_desc})",
                "min": 0,
                "modify": lambda x: x * config.read_count_multiplier,
                "scale": "Blues",
                "shared_key": "read_count",
            },
        }
        self.general_stats_addcols(self.seqwho_data, headers)

    def seqwho_species_plot(self):
        """
        Take the Maximum Likelihood Species data from the SeqWho report and make
        barchart
        """

        config = {
            "id": "seqwho_species_plot",
            "title": "SeqWho: Species",
            "ylab": "Maximum likelihood estimate",
            "ymax": 1.0,
            "stacking": None,
            "cpswitch": False,
            "cpswitch_counts_label": "Maximum likelihood",
            "tt_decimals": 3,
        }

        self.add_section(
            name="Species",
            anchor="seqwho_species",
            description="This plot shows the maximum likelihood that a given species matches.",
            plot=bargraph.plot(self.seqwho_data, SPECIES, config),
        )

    def seqwho_library_plot(self):
        """Take the Maximum Likelihood Library data from the SeqWho report and make a
        barplot"""

        # Config for the plot
        config = {
            "id": "seqwho_sequencing_plot",
            "title": "SeqWho: Library Type",
            "ylab": "Maximum likelihood estimate",
            "ymax": 1.0,
            "stacking": None,
            "cpswitch": False,
            "cpswitch_counts_label": "Maximum likelihood",
            "tt_decimals": 3,
        }

        self.add_section(
            name="Library Type",
            anchor="seqwho_library",
            description="This plot shows the maximum likelihood that a given library matches.",
            plot=bargraph.plot(self.seqwho_data, LIBRARIES, config),
        )

    def seqwho_qualdist_plot(self):
        """Generate the Average Sequence Quality Distribution plot"""
        pconfig = {
            "id": "seqwho_qualdis_plot",
            "title": "SeqWho: Per Sequence Quality Scores",
            "ylab": "Reads",
            "xlab": "Sequence Quality (Phred)",
            "xDecimals": False,
            "ymin": 0,
            "tt_label": "<b>Phred {point.x}</b>: {point.y:.0f} reads",
        }

        self.add_section(
            name="Predicted Per Sequence Quality Scores",
            anchor="seqwho_qualdis",
            description="This plot shows the predicted number of reads with average quality scores.",
            plot=linegraph.plot(self.seqwho_qualdis, pconfig),
        )

    def seqwho_qualscore_plot(self):
        """Generate the Average Quality Score per position plot"""
        pconfig = {
            "id": "seqwho_qualscore_plot",
            "title": "SeqWho: Quality Score",
            "ylab": "Sequence Quality (Phred)",
            "xlab": "Position (bp)",
            "xDecimals": False,
            "ymin": 0,
            "tt_label": "<b>Base {point.x} </b>: {point.y}",
        }

        self.add_section(
            name="Predicted Mean Quality Scores",
            anchor="seqwho_qualscore",
            description="This plot shows the predicted mean quality value across each base position in the read.",
            plot=linegraph.plot(self.seqwho_qualscore, pconfig),
        )

    def seqwho_readdist_plot(self):
        """Generate the number of reads with certain length plot"""
        pconfig = {
            "id": "seqwho_readdist_plot",
            "title": "SeqWho: Read Distribution",
            "ylab": "Reads",
            "xlab": "Length (bp)",
            "xDecimals": False,
            "ymin": 0,
            "tt_label": "<b>Base {point.x} bp</b>: {point.y} reads",
        }

        self.add_section(
            name="Predicted Read Distribution",
            anchor="seqwho_readdist",
            description="This plot shows predicted the number of reads with certain lengths.",
            plot=linegraph.plot(self.seqwho_readdist, pconfig),
        )



#!/usr/bin/env python

""" MultiQC functions to plot a scatter plot """

import base64
import io
import logging
import os
import random
import sys

from multiqc.utils import config, report

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(data, pconfig=None):
    """Plot a box-and-whisker plot
    :param data: 2D dict, first keys as read positions, then as quantile:QV pairs
    :param pconfig: optional dict with config key:value pairs. See CONTRIBUTING.md
    :return: HTML and JS, ready to be inserted into the page
    """
    if pconfig is None:
        pconfig = {}

    # Make a plot
    return matplotlib_boxplot(data, pconfig)


def mock_data(data):
    res = []
    value = 2  # Default to our minimum/ambiguous base QV
    for key in [1, 2, 5, 10, 25, 50, 75, 90]:
        value = data.get(key, value)
        if key in [10, 90]:
            res += [value] * 10
        elif key in [25, 75]:
            res += [value] * 25
        elif key == 50:
            res += [value] * 30
    return res


def mock_dataset(data):
    res = []
    for key in sorted(data.keys()):
        res.append(mock_data(data[key]))
    return res


def matplotlib_boxplot(plotdata, pconfig=None):
    """
    Plot a line graph with Matplot lib and return a HTML string. Either embeds a base64
    encoded image within HTML or writes the plot and links to it. Should be called by
    plot_bargraph, which properly formats the input data.
    """
    try:
        # Import matplot lib but avoid default X environment
        import matplotlib

        matplotlib.use("Agg")
        import matplotlib.pyplot as plt

        logger.debug(f"Using matplotlib version {matplotlib.__version__}")
    except Exception as e:
        # MatPlotLib can break in a variety of ways. Fake an error message and continue without it if so.
        # The lack of the library will be handled when plots are attempted
        print("##### ERROR! MatPlotLib library could not be loaded!    #####", file=sys.stderr)
        print("##### Flat plots will instead be plotted as interactive #####", file=sys.stderr)
        print(e)

    if pconfig is None:
        pconfig = {}

    # Plot group ID
    if pconfig.get("id") is None:
        pconfig["id"] = "mqc_mplplot_" + "".join(random.sample(letters, 10))

    # Sanitise plot ID and check for duplicates
    pconfig["id"] = report.save_htmlid(pconfig["id"])

    # Individual plot IDs
    pids = []
    for k in range(len(plotdata)):
        try:
            name = pconfig["data_labels"][k]["name"]
        except Exception:
            name = k + 1
        pid = f"mqc_{pconfig['id']}_{name}"
        pid = report.save_htmlid(pid, skiplint=True)
        pids.append(pid)

    html = (
        '<p class="text-info"><small><span class="glyphicon glyphicon-picture" aria-hidden="true"></span> '
        + "Flat image plot. Toolbox functions such as highlighting / hiding samples will not work "
        + '(see the <a href="http://multiqc.info/docs/#flat--interactive-plots" target="_blank">docs</a>).</small></p>'
    )
    html += f"<div class=\"mqc_mplplot_plotgroup\" id=\"{pconfig['id']}\">"

    # Buttons to cycle through different datasets
    if len(plotdata) > 1 and not config.simple_output:
        html += '<div class="btn-group mpl_switch_group mqc_mplplot_bargraph_switchds">\n'
        for k, p in enumerate(plotdata):
            pid = pids[k]
            active = "active" if k == 0 else ""
            try:
                name = pconfig["data_labels"][k]["name"]
            except Exception:
                name = k + 1
            html += '<button class="btn btn-default btn-sm {a}" data-target="#{pid}">{n}</button>\n'.format(
                a=active, pid=pid, n=name
            )
        html += "</div>\n\n"

    # Go through datasets creating plots
    for pidx, (pname, pdata) in enumerate(plotdata.items()):
        # Plot ID
        pid = pids[pidx]

        # Set up figure
        fig = plt.figure(figsize=(14, 6), frameon=False)
        axes = fig.add_subplot(111)
        plt.xticks(rotation=90)

        # Go through data series
        mock_ds = mock_dataset(pdata)
        box = axes.boxplot(mock_ds, whis=[0, 100], patch_artist=True)

        for patch in box["boxes"]:
            patch.set_facecolor("yellow")

        # Axis limits
        default_ylimits = axes.get_ylim()
        ymin = default_ylimits[0]
        if "ymin" in pconfig:
            ymin = pconfig["ymin"]
        elif "yFloor" in pconfig:
            ymin = max(pconfig["yFloor"], default_ylimits[0])
        ymax = default_ylimits[1]
        if "ymax" in pconfig:
            ymax = pconfig["ymax"]
        elif "yCeiling" in pconfig:
            ymax = min(pconfig["yCeiling"], default_ylimits[1])
        if (ymax - ymin) < pconfig.get("yMinRange", 0):
            ymax = ymin + pconfig["yMinRange"]
        axes.set_ylim((ymin, ymax))

        # Dataset specific ymax
        try:
            axes.set_ylim((ymin, pconfig["data_labels"][pidx]["ymax"]))
        except Exception:
            pass

        default_xlimits = axes.get_xlim()
        xmin = default_xlimits[0]
        if "xmin" in pconfig:
            xmin = pconfig["xmin"]
        elif "xFloor" in pconfig:
            xmin = max(pconfig["xFloor"], default_xlimits[0])
        xmax = default_xlimits[1]
        if "xmax" in pconfig:
            xmax = pconfig["xmax"]
        elif "xCeiling" in pconfig:
            xmax = min(pconfig["xCeiling"], default_xlimits[1])
        if (xmax - xmin) < pconfig.get("xMinRange", 0):
            xmax = xmin + pconfig.get("xMinRange", 0)
        axes.set_xlim((xmin, xmax))

        # Plot title
        if "title" in pconfig:
            if len(plotdata) > 1:
                title = f"{pconfig['title']} for {pname}"
            else:
                title = pconfig["title"]
            plt.text(0.5, 1.05, title, horizontalalignment="center", fontsize=16, transform=axes.transAxes)
        axes.set_xlabel(pconfig.get("xlab", ""))
        axes.set_ylabel(pconfig.get("ylab", ""))
        axes.grid(True, zorder=10, which="both", axis="y", linestyle="-", color="#dedede", linewidth=1)

        # X axis categories, if specified
        if "categories" in pconfig:
            axes.set_xticks([i for i, v in enumerate(pconfig["categories"])])
            axes.set_xticklabels(pconfig["categories"])

        # Axis lines
        xlim = axes.get_xlim()
        axes.plot([xlim[0], xlim[1]], [0, 0], linestyle="-", color="#dedede", linewidth=2)
        axes.set_axisbelow(True)
        axes.spines["right"].set_visible(False)
        axes.spines["top"].set_visible(False)
        axes.spines["bottom"].set_visible(False)
        axes.spines["left"].set_visible(False)

        # Background colours, if specified
        if "yPlotBands" in pconfig:
            xlim = axes.get_xlim()
            for pb in pconfig["yPlotBands"]:
                axes.barh(
                    pb["from"],
                    xlim[1],
                    height=pb["to"] - pb["from"],
                    left=xlim[0],
                    color=pb["color"],
                    linewidth=0,
                    zorder=0,
                    align="edge",
                )
        if "xPlotBands" in pconfig:
            ylim = axes.get_ylim()
            for pb in pconfig["xPlotBands"]:
                axes.bar(
                    pb["from"],
                    ylim[1],
                    width=pb["to"] - pb["from"],
                    bottom=ylim[0],
                    color=pb["color"],
                    linewidth=0,
                    zorder=0,
                    align="edge",
                )

        # Tight layout - makes sure that legend fits in and stuff
        if len(pdata) <= 15:
            axes.legend(
                loc="lower center",
                bbox_to_anchor=(0, -0.22, 1, 0.102),
                ncol=5,
                mode="expand",
                fontsize=8,
                frameon=False,
            )
            plt.tight_layout(rect=[0, 0.08, 1, 0.92])
        else:
            plt.tight_layout(rect=[0, 0, 1, 0.92])

        # Should this plot be hidden on report load?
        hidediv = ""
        if pidx > 0:
            hidediv = ' style="display:none;"'

        # Save the plot to the data directory if export is requests
        if config.export_plots:
            for fformat in config.export_plot_formats:
                # Make the directory if it doesn't already exist
                plot_dir = os.path.join(config.plots_dir, fformat)
                if not os.path.exists(plot_dir):
                    os.makedirs(plot_dir)
                # Save the plot
                plot_fn = os.path.join(plot_dir, f"{pid}.{fformat}")
                fig.savefig(plot_fn, format=fformat, bbox_inches="tight")

        # Output the figure to a base64 encoded string
        if getattr(get_template_mod(), "base64_plots", True) is True:
            img_buffer = io.BytesIO()
            fig.savefig(img_buffer, format="png", bbox_inches="tight")
            b64_img = base64.b64encode(img_buffer.getvalue()).decode("utf8")
            img_buffer.close()
            html += '<div class="mqc_mplplot" id="{}"{}><img src="data:image/png;base64,{}" /></div>'.format(
                pid, hidediv, b64_img
            )

        # Save to a file and link <img>
        else:
            plot_relpath = os.path.join(config.plots_dir_name, "png", f"{pid}.png")
            html += f'<div class="mqc_mplplot" id="{pid}"{hidediv}><img src="{plot_relpath}" /></div>'

        plt.close(fig)

    # Close wrapping div
    html += "</div>"

    report.num_mpl_plots += 1

    return html



""" MultiQC module to parse Stacks 2 denovo output"""


import logging
import os
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="Stacks",
            anchor="stacks",
            href="http://catchenlab.life.illinois.edu/stacks/",
            info="A software for analyzing restriction enzyme-based data (e.g. RAD-seq).",
            doi="10.1111/mec.12354",
        )

        gsheaders = {
            "n_loci": {
                "title": "# loci",
                "description": "Number of loci built",
                "format": "{:,.i}",
                "scale": "RdYlGn",
            },
            "n_used_fw_reads": {
                "title": "K reads used",
                "modify": lambda x: float(x) / 1000.0,
                "description": "Number of thousand reads used",
                "scale": "BuGn",
            },
            "mean_cov": {
                "title": "cov",
                "suffix": "X",
                "description": "Mean sequence coverage at locus",
                "scale": "BuPu",
            },
            "mean_cov_ns": {
                "title": "weighted cov",
                "suffix": "X",
                "description": "The coverage at each locus is weighted by the number of samples present at that locus (i.e. coverage at shared loci counts more)",
                "scale": "YlGn",
            },
        }

        sheaders = {
            "# Pop ID": {
                "title": "PopID",
                "description": "Population ID as defined in the Population Map file.",
                "scale": False,
                "format": "{:,.s}",
            },
            "Private": {
                "title": "Private",
                "description": "Number of private alleles in this population.",
                "scale": "PuBu",
                "hidden": True,
            },
            "Num_Indv": {
                "title": "# Indv",
                "description": "Mean number of individuals per locus in this population.",
                "scale": "YlGn",
            },
            "P": {
                "title": "P",
                "description": "Mean frequency of the most frequent allele at each locus in this population.",
                "scale": "PuBu",
                "min": 0,
                "max": 1,
            },
            "Obs_Het": {
                "title": "Obs Het",
                "description": "Mean observed heterozygosity in this population.",
                "scale": "YlGn",
                "min": 0,
                "max": 1,
            },
            "Obs_Hom": {
                "title": "Obs Hom",
                "description": "Mean observed homozygosity in this population.",
                "scale": "PuBu",
                "min": 0,
                "max": 1,
                "hidden": True,
            },
            "Exp_Hom": {
                "title": "Exp_Hom",
                "description": "Mean expected homozygosity in this population.",
                "scale": "YlGn",
                "min": 0,
                "max": 1,
                "hidden": True,
            },
            "Exp_Het": {
                "title": "Exp Het",
                "description": "Mean expected heterozygosity in this population.",
                "scale": "PuBu",
                "min": 0,
                "max": 1,
            },
            "Pi": {
                "title": "Pi",
                "description": "Mean value of &#960; in this population.",
                "scale": "YlGn",
                "min": 0,
                "max": 1,
            },
            "Fis": {
                "title": "Fis",
                "description": "Mean measure of Fis in this population.",
                "scale": "PuOr",
                "min": -1,
                "max": 1,
            },
        }

        num_files = 0
        # Parse gstacks data
        cov_data = {}
        for f in self.find_log_files("stacks/gstacks"):
            run_name = os.path.dirname(f["root"])
            s_name = self.clean_s_name(os.path.basename(f["root"]), f, root=run_name)
            try:
                cov_data.update(self.parse_gstacks(f["f"], s_name))
                self.add_data_source(f, section="gstacks")
                num_files += 1
            except Exception as e:
                log.error(f"Could not parse gstacks.distribs file in {f['s_name']}: {e}")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Parse populations data
        distribs_loci = {}
        distribs_snps = {}
        for f in self.find_log_files("stacks/populations"):
            run_name = os.path.dirname(f["root"])
            s_name = self.clean_s_name(os.path.basename(f["root"]), f, root=run_name)
            i, j = self.parse_populations(f["f"], s_name)
            try:
                distribs_loci.update(i)
                distribs_snps.update(j)
                self.add_data_source(f, section="populations")
                num_files += 1
            except Exception as e:
                log.error(f"Could not parse population.log.distribs file in {f['s_name']}: {e}")

        # Parse sumstats file
        sumstats_data = {}
        for f in self.find_log_files("stacks/sumstats"):
            run_name = os.path.dirname(f["root"])
            s_name = self.clean_s_name(os.path.basename(f["root"]), f, root=run_name)
            try:
                sumstats_data.update(self.parse_sumstats(f["f"], s_name))
                self.add_data_source(f, section="sumstats")
                num_files += 1
            except Exception as e:
                log.error(f"Could not parse populations.sumstats_summary file in {f['s_name']}: {e}")

        # Ignore samples
        cov_data = self.ignore_samples(cov_data)
        distribs_loci = self.ignore_samples(distribs_loci)
        distribs_snps = self.ignore_samples(distribs_snps)
        sumstats_data = self.ignore_samples(sumstats_data)

        if len(cov_data) == 0 and len(sumstats_data) == 0 and len(distribs_loci) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {num_files} reports")

        # Write parsed report data to a file
        self.write_data_file(cov_data, "multiqc_stacks_cov")
        self.write_data_file(sumstats_data, "multiqc_stacks_sumstats")

        # Write the sample table
        config_table = {
            "id": "gstacks_table",
            "namespace": "stacks",
            "title": "Stacks: Sample statistics",
        }
        self.add_section(
            name="Sample statistics",
            anchor="stacks-gstacks",
            description="The sample specific statistics for Stacks",
            helptext="""**Note!** The sample names have the following scheme `<run folder name> | <input fastq file prefix>`.
                        This data is obtained from the gstacks program run after builing sample and catalog loci merge
                        paired-ends and call variants.
                        These numbers are obtained from the `gstacks.log.distribs` file""",
            plot=table.plot(cov_data, gsheaders, config_table),
        )
        # Write population sumstats table
        config_table = {
            "id": "sumstats_table",
            "namespace": "stacks",
            "title": "Stacks: Population summary statistics",
        }
        self.add_section(
            name="Population summary statistics",
            anchor="stacks-sumstats",
            description="Population statistics as calculated from variant sites found in this run",
            helptext="""**Note!** The sample names have the following scheme `<run folder name> | <population ID>`,
                        where the population ID is defined in the input population map file.
                        This information is obtained from the Stacks program `population` and the file populations.sumstats_summary.tsv
                        """,
            plot=table.plot(sumstats_data, sheaders, config_table),
        )
        config_distribs = {
            "id": "distribs_plot",
            "title": "Stacks: Population plots",
            "namespace": "stacks",
            "tt_label": "{point.y} loci, {point.x} samples/SNPs",
            "ylab": "# loci",
            "data_labels": [
                {"name": "Samples per loci", "ylab": "# loci", "xlab": "# samples"},
                {"name": "SNPs per loci", "ylab": "# loci", "xlab": "# SNPs"},
            ],
        }
        self.add_section(
            name="Population plots",
            anchor="stacks-distribs",
            description="Plots showing, 1) the number of loci shared by number of samples and 2) the number of SNPs per sample",
            helptext="""The distributions are obtained from the Stacks program `populations` and it's output file `populations.log.distribs`.
            These numbers are Stacks' post-filtering.""",
            plot=linegraph.plot([distribs_loci, distribs_snps], config_distribs),
        )

    @staticmethod
    def parse_gstacks(file_contents, s_name):
        headers = None
        out_dict = {}

        for line in file_contents.splitlines():
            if line.startswith("sample	n_loci	n_used_fw_reads"):
                headers = ["n_loci", "n_used_fw_reads", "mean_cov", "mean_cov_ns"]
            elif line.startswith("END effective_coverages_per_sample"):
                break
            elif headers is not None:
                cdict = dict.fromkeys(headers)
                content = line.split("\t")
                for i in range(1, len(content)):
                    cdict[list(cdict.keys())[i - 1]] = content[i]
                if s_name is not None:
                    out_dict[s_name + " | " + content[0]] = cdict
                else:
                    out_dict[content[0]] = cdict

        return out_dict

    @staticmethod
    def parse_sumstats(file_contents, s_name):
        out_dict = dict()
        # ["# Pop ID","Private","Num_Indv","P","Obs_Het","Obs_Hom","Exp_Het","Exp_Hom","Pi","Fis"]
        fields = [0, 1, 2, 5, 8, 11, 14, 17, 20, 23]
        fl = file_contents.splitlines()
        var_lines = fl[fl.index("# Variant positions") + 1 : fl.index("# All positions (variant and fixed)")]
        headers = None
        for line in var_lines:
            if line.startswith("# Pop ID	Private	Num_Indv"):
                headers = line.split("\t")
            elif headers is not None:
                cdict = {}
                content = line.split("\t")
                for i in fields:
                    cdict[headers[i]] = content[i]
                out_dict[s_name + " | " + content[0]] = cdict

        return out_dict

    @staticmethod
    def parse_populations(file_contents, s_name):
        loci_dict = {}
        snps_dict = {}
        pat = re.compile("BEGIN (.*)\n\n{0,1}#.*\n.+\n((?:.+\n)+)END", re.MULTILINE)
        for match in pat.finditer(file_contents):
            title, table = match.groups()
            title = title.strip()
            if title == "missing_samples_per_loc_postfilters":
                if s_name is not None:
                    title = s_name
                loci_dict[title] = {}
                for line in table.splitlines():
                    row = line.split()
                    loci_dict[title][int(row[0])] = int(row[1])
            elif title == "snps_per_loc_postfilters":
                if s_name is not None:
                    title = s_name
                snps_dict[title] = {}
                for line in table.splitlines():
                    row = line.split()
                    snps_dict[title][int(row[0])] = int(row[1])

        return loci_dict, snps_dict



import logging
import re
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class DragenCoveragePerContig(BaseMultiqcModule):
    def add_coverage_per_contig(self):
        perchrom_data_by_phenotype_by_sample = defaultdict(dict)

        for f in self.find_log_files("dragen/wgs_contig_mean_cov"):
            perchrom_data_by_phenotype = parse_wgs_contig_mean_cov(f)
            s_name = f["s_name"]
            if s_name in perchrom_data_by_phenotype_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, section="wgs_contig_mean_cov")
            perchrom_data_by_phenotype_by_sample[s_name].update(perchrom_data_by_phenotype)

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Filter to strip out ignored sample names:
        perchrom_data_by_phenotype_by_sample = self.ignore_samples(perchrom_data_by_phenotype_by_sample)

        # Merge tumor and normal data:
        perchrom_data_by_sample = defaultdict(dict)
        for sn in perchrom_data_by_phenotype_by_sample:
            for phenotype in perchrom_data_by_phenotype_by_sample[sn]:
                new_sn = sn
                if phenotype == "normal":
                    new_sn = sn + "_normal"
                perchrom_data_by_sample[new_sn] = perchrom_data_by_phenotype_by_sample[sn][phenotype]

        if not perchrom_data_by_sample:
            return set()

        # Data is in wrong format for writing to file
        # self.write_data_file(perchrom_data_by_sample, "dragen_cov_contig")

        main_contigs_by_sample = {sn: data[0] for sn, data in perchrom_data_by_sample.items()}
        other_contigs_by_sample = {sn: data[1] for sn, data in perchrom_data_by_sample.items()}

        self.add_section(
            name="Coverage per contig",
            anchor="dragen-coverage-per-contig",
            description="""
            Average coverage per contig or chromosome.
            Calculated as the number of bases (excluding duplicate marked reads, reads
            with MAPQ=0, and clipped bases), divided by the length of the contig or
            (if a target bed is used) the total length of the target region spanning that contig.
            """,
            plot=linegraph.plot(
                main_contigs_by_sample,
                pconfig={
                    "id": "dragen_coverage_per_contig",
                    "title": "Dragen: Average coverage per contig"
                    + (" (main contigs)" if other_contigs_by_sample else ""),
                    "ylab": "Average coverage",
                    "xlab": "Region",
                    "categories": True,
                    "tt_label": "<b>{point.x}</b>: {point.y:.1f}x",
                },
            ),
        )

        if other_contigs_by_sample:
            self.add_section(
                name="Coverage per contig (non-main)",
                anchor="dragen-coverage-per-nonmain-contig",
                description="""
                Non-main contigs:
                unlocalized (*_random), unplaced (chrU_*), alts (*_alt), mitochondria (chrM), EBV, HLA.
                Zoom in to see more contigs as all labels don\'t fit the screen.
                """,
                plot=linegraph.plot(
                    other_contigs_by_sample,
                    pconfig={
                        "id": "dragen_coverage_per_non_main_contig",
                        "title": "Dragen: Average coverage of non-main contigs",
                        "ylab": "Average coverage",
                        "xlab": "Region",
                        "categories": True,
                        "tt_label": "<b>{point.x}</b>: {point.y:.1f}x",
                    },
                ),
            )

        return perchrom_data_by_sample.keys()


def parse_wgs_contig_mean_cov(f):
    """
    The Contig Mean Coverage report generates a _contig_mean_cov.csv file, which contains the estimated coverage for
    all contigs, and an autosomal estimated coverage. The file includes the following three columns

    1. Contig name
    2. Number of bases aligned to that contig, which excludes bases from duplicate marked reads, reads with MAPQ=0,
       and clipped bases.
    3. Estimated coverage, as follows: <number of bases aligned to the contig (ie, Col2)> divided by <length of the
       contig or (if a target bed is used) the total length of the target region spanning that contig>.

    T_SRR7890936_50pc.wgs_contig_mean_cov_normal.csv
    T_SRR7890936_50pc.wgs_contig_mean_cov_tumor.csv

    chr1,11292297134,48.9945
    chr10,6482885699,48.6473
    ...
    chrUn_GL000218v1,20750824,128.77
    chrX,3590295769,23.1792
    chrY,42229820,1.5987
    chrY_KI270740v1_random,0,0
    Autosomal regions ,130912665915,47.4953

    A histogram or a plot like in mosdepth, with each chrom in X axis.
    Two versions: for main contigs (chr1..chrY), and including non-main contigs (chrUn_*, *_random, HLA-*, *_alt, chrM)
    """

    main_contig_perchrom_data = dict()
    other_contig_perchrom_data = dict()

    for line in f["f"].splitlines():
        chrom, bases, depth = line.split(",")
        chrom = chrom.strip()
        depth = float(depth)
        # skipping unplaced and alternative contigs, as well as the mitochondria (might attract 100 times more coverage
        # than human chromosomes):
        if (
            chrom.startswith("chrUn_")
            or chrom.endswith("_random")
            or chrom.endswith("_alt")
            or chrom == "chrM"
            or chrom == "MT"
            or chrom == "chrEBV"
            or chrom.startswith("HLA-")
        ):
            other_contig_perchrom_data[chrom] = depth
        else:
            main_contig_perchrom_data[chrom] = depth

    def chrom_order(chrom):
        if chrom == "Autosomal regions":
            # "Autosomal regions" average coverage goes right after all the autosomal chromosomes
            return 0
        try:
            # autosomal chromosomes go first, thus getting a negative order
            return int(chrom.replace("chr", "")) - len(main_contig_perchrom_data)
        except ValueError:
            # sex and other chromosomes go in the end
            return 1

    main_contig_perchrom_data = dict(
        sorted(
            main_contig_perchrom_data.items(),
            key=lambda key_val: chrom_order(key_val[0]),
        )
    )
    other_contig_perchrom_data = dict(
        sorted(
            other_contig_perchrom_data.items(),
            key=lambda key_val: chrom_order(key_val[0]),
        )
    )

    m = re.search(r"(tumor|normal).csv", f["fn"])
    if m:
        phenotype = m.group(1)
    else:
        phenotype = "unknown"
    return {phenotype: [main_contig_perchrom_data, other_contig_perchrom_data]}



""" MultiQC module to parse output from Space Ranger """

import logging

from multiqc.modules.base_module import BaseMultiqcModule

# Import the Samtools submodules
from .count import SpaceRangerCountMixin

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule, SpaceRangerCountMixin):
    """Space Ranger has 2 main modules: count and vdj.
    This module parses data directly from the web_summary.html and summarise
    data useful for QC in the main sample table as well as generate some module specific plots
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Space Ranger",
            anchor="spaceranger",
            href="https://support.10xgenomics.com/spatial-gene-expression/software/pipelines/latest/what-is-space-ranger",
            info="Space Ranger analyzes 10x Genomics Visium spatial transcriptomics data.",
            doi=[],
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = dict()
        self.general_stats_data = dict()
        n = dict()

        # Call submodule functions
        n["count"] = self.parse_count_html()
        if n["count"] > 0:
            log.info("Found {} Space Ranger count reports".format(n["count"]))

        # Exit if we didn't find anything
        if sum(n.values()) == 0:
            raise UserWarning

        # Add to the General Stats table (has to be called once per MultiQC module)
        self.general_stats_addcols(self.general_stats_data, self.general_stats_headers)



""" MultiQC module to parse output from Snippy """

import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    snippy module class.
    """

    # -------------------------------------------------------------------------#
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Snippy",
            anchor="snippy",
            href="https://github.com/tseemann/snippy",
            info="is used for rapid haploid variant calling and core genome alignment.",
            # Can't find a DOI // doi=
        )

        self.snippy_data = {}
        self.snippy_core_data = {}
        self.snippy_col = [
            "Variant-COMPLEX",
            "Variant-DEL",
            "Variant-INS",
            "Variant-SNP",
            "VariantTotal",
        ]
        self.snippy_core_col = [
            "LENGTH",
            "ALIGNED",
            "UNALIGNED",
            "VARIANT",
            "HET",
            "MASKED",
            "LOWCOV",
        ]

        # Parse the txt files from snippy
        for f in self.find_log_files("snippy/snippy"):
            # Check for duplicate sample names
            if f["s_name"] in self.snippy_data:
                log.debug(f"Duplicate sample name found for snippy! Overwriting: {f['s_name']}")
            # Add the file data under the key filename
            data = self.parse_snippy_txt(f["f"])
            if data:
                self.snippy_data[f["s_name"]] = data
                self.add_software_version(data["version"], f["s_name"])

            self.add_data_source(f, section="snippy")

        # Ignore samples specified by the user
        self.snippy_data = self.ignore_samples(self.snippy_data)

        # Parse the txt files from snippy-core
        for f in self.find_log_files("snippy/snippy-core"):
            # Check for duplicate sample names
            if f["s_name"] in self.snippy_core_data:
                log.debug(f"Duplicate sample name found for snippy-core! Overwriting: {f['s_name']}")
            # Add the file data under the key filename
            self.snippy_core_data[f["s_name"]] = self.parse_snippy_core_txt(f["f"])

            self.add_data_source(f, section="snippy-core")

        # Ignore samples specified by the user
        self.snippy_core_data = self.ignore_samples(self.snippy_core_data)

        # Raise warning if no logs were found.
        if len(self.snippy_data) == 0 and len(self.snippy_core_data) == 0:
            raise ModuleNoSamplesFound

        # Run analysis if txt files found
        if len(self.snippy_data) > 0:
            log.info(f"Found {len(self.snippy_data)} reports")
            self.snippy_stats_table()
            self.snippy_report_section()

        if len(self.snippy_core_data) > 0:
            log.info(f"Found {len(self.snippy_core_data)} snippy-core reports")
            self.snippy_core_stats_table()
            self.snippy_core_report_section()

    # -------------------------------------------------------------------------#
    def parse_snippy_txt(self, file):
        """
        Parse the txt file for snippy output.
        """
        data = {}
        for line in file.splitlines():
            split_line = line.strip().split("\t")
            if split_line[0] in self.snippy_col:
                data[split_line[0]] = int(split_line[1])

            if split_line[0] == "Software":
                data["version"] = split_line[1].split(" ")[1]

        if len(data) == 0:
            return False
        for col in self.snippy_col:
            if col not in data:
                data[col] = 0
        return data

    # -------------------------------------------------------------------------#
    def parse_snippy_core_txt(self, file):
        """
        Parse the txt file for snippy-core output.
        """
        data = {}
        for line in file.splitlines()[1:]:
            split_line = line.split("\t")
            data[split_line[0]] = split_line[1:]
        return data

    # -------------------------------------------------------------------------#
    def snippy_stats_table(self):
        """
        Add the snippy data to the general stats.
        """
        self.general_stats_addcols(
            self.snippy_data,
            {
                "VariantTotal": {
                    "title": "# Variants",
                    "description": "Total variants detected.",
                    "scale": "BuPu",
                    "format": "{:.0f}",
                    "shared_key": "variant_count",
                }
            },
        )
        self.write_data_file(self.snippy_data, "multiqc_snippy")

    # -------------------------------------------------------------------------#
    def snippy_core_stats_table(self):
        """
        Add the snippy-core data to the general stats.
        """
        # Parse the statistics
        data = {}
        for file in self.snippy_core_data:
            for sample in self.snippy_core_data[file]:
                data[sample] = {}
                for i in range(0, len(self.snippy_core_col)):
                    data[sample][self.snippy_core_col[i]] = int(self.snippy_core_data[file][sample][i])
                data[sample]["Percent_Aligned"] = (data[sample]["ALIGNED"] / data[sample]["LENGTH"]) * 100
                data[sample]["Percent_Het"] = (data[sample]["HET"] / data[sample]["ALIGNED"]) * 100

        self.general_stats_addcols(data, self.snippy_core_headers_config())
        self.write_data_file(data, "multiqc_snippy")

    # -------------------------------------------------------------------------#
    def snippy_report_section(self):
        """
        Create a report section for the snippy data.
        """
        bargraph_data = {}
        for sample in self.snippy_data:
            # Remove the VariantTotal stat
            bargraph_data[sample] = {}
            for stat in self.snippy_data[sample]:
                if stat == "VariantTotal":
                    continue
                bargraph_data[sample][stat] = self.snippy_data[sample][stat]

        # Config for the plot
        pconfig = {
            "id": "snippy_variants",
            "title": "Snippy: Variants Counts",
            "ylab": "# Variants",
        }

        self.add_section(
            name="Snippy Variants",
            anchor="snippy_variants_stats",
            plot=bargraph.plot(bargraph_data, pconfig=pconfig),
            description="This plot shows the different variant types reported by snippy.",
        )

    # -------------------------------------------------------------------------#
    def snippy_core_report_section(self):
        """
        Create a report section for the snippy-core data.
        """
        bargraph_data = {}
        for file in self.snippy_core_data:
            for sample in self.snippy_core_data[file]:
                bargraph_data[sample] = {}
                for i in range(0, len(self.snippy_core_col)):
                    if self.snippy_core_col[i] == "LENGTH":
                        continue
                    bargraph_data[sample][self.snippy_core_col[i]] = int(self.snippy_core_data[file][sample][i])
                # Subtract variant from aligned
                bargraph_data[sample]["ALIGNED"] = bargraph_data[sample]["ALIGNED"] - bargraph_data[sample]["VARIANT"]

        # Config for the plot
        pconfig = {
            "id": "snippy_core_alignment",
            "title": "Snippy: Core Alignment Statistics",
            "ylab": "# bp",
        }

        self.add_section(
            name="Snippy-Core Alignment Statistics",
            anchor="snippy_core_alignment_stats",
            plot=bargraph.plot(bargraph_data, pconfig=pconfig),
            description="Different categories of sites in a snippy-core genome alignment.",
        )

    # -------------------------------------------------------------------------#
    def snippy_core_headers_config(self):
        """
        Prepare the headers for the snippy core stats table.
        """
        # General stats table headers
        headers = {
            "Percent_Het": {
                "title": "% Het",
                "description": "Percent of aligned sites that are heterozygous",
                "max": 100,
                "min": 0,
                "scale": "RdYlGn-rev",
                "suffix": "%",
            },
            "HET": {
                "title": "# Hets",
                "description": "Number of heterozygous sites",
                "scale": "RdYlGn-rev",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "VARIANT": {
                "title": "# Variants",
                "description": "Number of variants",
                "scale": "Blues",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "Percent_Aligned": {
                "title": "% Aligned",
                "description": "Percent of bases aligned to the reference",
                "max": 100,
                "min": 0,
                "scale": "RdYlGn",
                "suffix": "%",
            },
            "ALIGNED": {
                "title": "# Aligned",
                "description": "Number of aligned nucleotides",
                "scale": "RdYlGn",
                "format": "{:,.0f}",
                "hidden": True,
            },
        }
        return headers



""" MultiQC module to parse output from QualiMap """


import logging
import os
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """Qualimap is really a collection of separate programs:
    BamQC, RNASeq and Counts.. This module is split into separate
    files to reflect this and help with code organisation."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="QualiMap",
            anchor="qualimap",
            href="http://qualimap.bioinfo.cipf.es/",
            info="is a platform-independent application to facilitate the quality"
            " control of alignment sequencing data and its derivatives like"
            " feature counts.",
            doi=["10.1093/bioinformatics/btv566", "10.1093/bioinformatics/bts503"],
        )

        # Initialise the submodules
        from . import QM_BamQC, QM_RNASeq

        # Set up class objects to hold parsed data()
        self.general_stats_headers = dict()
        self.general_stats_data = defaultdict(lambda: dict())
        n = dict()

        # Call submodule functions
        n["BamQC"] = QM_BamQC.parse_reports(self)
        if n["BamQC"] > 0:
            log.info(f"Found {n['BamQC']} BamQC reports")

        n["RNASeq"] = QM_RNASeq.parse_reports(self)
        if n["RNASeq"] > 0:
            log.info(f"Found {n['RNASeq']} RNASeq reports")

        # Exit if we didn't find anything
        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound

        # Remove filtered samples from general stats table
        self.general_stats_data = self.ignore_samples(self.general_stats_data)

        # Add to the General Stats table (has to be called once per MultiQC module)
        self.general_stats_addcols(self.general_stats_data, self.general_stats_headers)

    # Helper functions
    def get_s_name(self, f):
        s_name = os.path.basename(os.path.dirname(f["root"]))
        s_name = self.clean_s_name(s_name, f)
        if s_name.endswith(".qc"):
            s_name = s_name[:-3]
        return s_name
